#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os

from io import StringIO

from pysmt.shortcuts import reset_env
from pysmt.test import TestCase
from pysmt.smtlib.parser import SmtLibParser

class TestSmtLibParserGriggio(TestCase):

    def test_griggio(self):
        for file_id in range(1,7):
            # Parse
            script = self.parse(file_id)
            # Check cmds
            for i, cmd in enumerate(script):
                self.assertEqual(cmd.name, TESTS[file_id][i],
                                 "Test %d: %s != %s " %
                                 (file_id, cmd.name, TESTS[file_id][i]))
            # Serialize
            buf = StringIO()
            script.serialize(buf)
            self.assertTrue(True)

    def parse(self, file_id):
        fname = SMTLIB_FILE_PATTERN % file_id
        reset_env()
        parser = SmtLibParser()
        script = parser.get_script_fname(fname)
        self.assertIsNotNone(script)
        return script


SMTLIB_FILE_PATTERN = "pysmt/test/smtlib/griggio/test%d.smt2.bz2"

TESTS = {

    1 : ["declare-fun",
         "set-option",
         "declare-sort",
         "declare-fun",
         "declare-fun",
         "assert",
         "assert",
         "push",
         "define-fun",
         "assert",
         "check-sat",
         "pop",
         "define-fun",
         "exit",
    ],

    2 : ["push",
         "declare-sort",
         "pop",
         "declare-fun",
         "exit",
    ],

    3 : [ "set-option",
          "declare-fun",
          "declare-fun",
          "assert",
          "assert",
          "assert",
          "check-sat",
          "get-unsat-core",
          "exit",
    ],

    4 : [ "set-option",
         "declare-fun",
         "declare-fun",
         "assert",
         "assert",
         "assert",
         "check-sat",
         "get-assignment",
         "exit",
    ],

    5 : ["set-option",
         "declare-fun",
         "declare-fun",
         "declare-fun",
         "assert",
         "assert",
         "assert",
         "assert",
         "check-sat",
         "get-value",
         "exit",
    ],

    6 : ["declare-sort",
         "define-sort",
         "declare-fun",
         "declare-fun",
         "assert",
         "check-sat",
         "exit",
    ]
}



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from pysmt.typing import BOOL
from pysmt.solvers.options import SolverOptions
from pysmt.decorators import clear_pending_pop
from pysmt.exceptions import (SolverReturnedUnknownResultError, PysmtValueError,
                              SolverNotConfiguredForUnsatCoresError,
                              PysmtTypeError, SolverStatusError)


class Solver(object):
    """Represents a generic SMT Solver."""

    # Define the supported logics for the Solver
    LOGICS = []

    # Class defining options for the Solver
    OptionsClass = SolverOptions

    def __init__(self, environment, logic, **options):
        if logic is None:
            raise PysmtValueError("Cannot provide 'None' as logic")

        self.environment = environment
        self.pending_pop = False
        self.logic = logic
        self.options = self.OptionsClass(**options)
        self._destroyed = False
        return

    def solve(self, assumptions=None):
        """Returns the satisfiability value of the asserted formulas.

        Assumptions is a list of Boolean variables or negations of
        boolean variables. If assumptions is specified, the
        satisfiability result is computed assuming that all the
        specified literals are True.

        A call to solve([a1, ..., an]) is functionally equivalent to::

          push()
          add_assertion(And(a1, ..., an))
          res = solve()
          pop()
          return res

        but is in general more efficient.

        Other convenience methods (is_sat, is_unsat, is_valid) are
        wrappers around this function.

        :returns: Whether the assertion stack is satisfiable
                  w.r.t. the given assumptions (if given)
        :rtype: bool
        """
        raise NotImplementedError

    def get_model(self):
        """Returns an instance of Model that survives the solver instance.

        Restrictions: Requires option generate_models to be set to
                      true (default) and can be called only after
                      :py:func:`solve` (or one of the derived methods)
                      returned sat or unknown, if no change to the
                      assertion set occurred.

        """
        raise NotImplementedError


    def is_sat(self, formula):
        """Checks satisfiability of the formula w.r.t. the current state of
        the solver.

        Previous assertions are taken into account.

        :type formula: FNode
        :returns: Whether formula is satisfiable
        :rtype: bool
        """
        assert formula in self.environment.formula_manager, \
               "Formula does not belong to the current Formula Manager"

        if not self.options.incremental:
            # If not incremental, we only need to assert and solve.
            self.add_assertion(formula)
            # We do not allow two calls to solve()
            def solve_error(*args, **kwargs):
                raise SolverStatusError("Cannot call is_sat twice when incrementality is disable")
            res = self.solve()
            self.solve = solve_error
            self.is_sat = solve_error
            return res

        # Try to be incremental using push/pop but fallback to
        # solving under assumption if push/pop is not implemented
        use_solving_under_assumption = False
        try:
            self.push()
        except NotImplementedError:
            use_solving_under_assumption = True

        if use_solving_under_assumption:
            res = self.solve([formula])
        else:
            self.add_assertion(formula)
            res = self.solve()
            self.pending_pop = True

        return res

    def is_valid(self, formula):
        """Checks validity of the formula w.r.t. the current state of the
        solver.

        Previous assertions are taken into account. See :py:func:`is_sat`

        :type formula: FNode
        :returns: Whether formula is valid
        :rtype: bool
        """
        Not = self.environment.formula_manager.Not
        return not self.is_sat(Not(formula))

    def is_unsat(self, formula):
        """Checks unsatisfiability of the formula w.r.t. the current state of
        the solver.

        Previous assertions are taken into account. See :py:func:`is_sat`

        :type formula: FNode
        :returns: Whether formula is unsatisfiable
        :rtype: bool
        """
        return not self.is_sat(formula)

    def get_values(self, formulae):
        """Returns the value of the expressions if a model was found.

        Requires option generate_models to be set to true (default)
        and can be called only after :py:func:`solve` (or to one of
        the derived methods) returned sat or unknown, if no change to
        the assertion set occurred.

        :type formulae: Iterable of FNodes
        :returns: A dictionary associating to each expr a value
        :rtype: dict

        """
        res = {}
        for f in formulae:
            v = self.get_value(f)
            res[f] = v
        return res

    def push(self, levels=1):
        """Push the current context of the given number of levels.

        :type levels: int
        """
        raise NotImplementedError

    def pop(self, levels=1):
        """Pop the context of the given number of levels.

        :type levels: int
        """
        raise NotImplementedError

    def exit(self):
        """Exits from the solver and closes associated resources."""
        if not self._destroyed:
            self._exit()
            self._destroyed = True

    def _exit(self):
        """Exits from the solver and closes associated resources."""
        raise NotImplementedError

    def reset_assertions(self):
        """Removes all defined assertions."""
        raise NotImplementedError

    def add_assertion(self, formula, named=None):
        """Add assertion to the solver."""
        raise NotImplementedError

    def add_assertions(self, formulae):
        for formula in formulae:
            self.add_assertion(formula)

    def print_model(self, name_filter=None):
        """Prints the model (if one exists).

        An optional function can be passed, that will be called on each symbol
        to decide whether to print it.
        """
        raise NotImplementedError

    def get_value(self, formula):
        """Returns the value of formula in the current model (if one exists).

        This is a simplified version of the SMT-LIB function get_values
        """
        raise NotImplementedError

    def get_py_value(self, formula):
        """Returns the value of formula as a python type.

        E.g., Bool(True) is translated into True.
        This simplifies writing code that branches on values in the model.
        """
        res = self.get_value(formula)
        assert res.is_constant()
        return res.constant_value()

    def get_py_values(self, formulae):
        """Returns the values of the formulae as python types.

        Returns a dictionary mapping each formula to its python value.
        """
        res = {}
        for f in formulae:
            v = self.get_py_value(f)
            res[f] = v
        return res

    def __enter__(self):
        """Manages entering a Context (i.e., with statement)"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Manages exiting from Context (i.e., with statement)

        The default behaviour is "close" the solver by calling the
        py:func:`exit` method.
        """
        self.exit()

    def _assert_no_function_type(self, item):
        """Enforces that argument 'item' cannot be a FunctionType.

        Raises TypeError.
        """
        if item.is_symbol() and item.symbol_type().is_function_type():
            raise PysmtTypeError("Cannot call get_value() on a FunctionType")

    def _assert_is_boolean(self, formula):
        """Enforces that argument 'formula' is of type Boolean.

        Raises TypeError.
        """
        if formula.get_type() != BOOL:
            raise PysmtTypeError("Argument must be boolean.")


class IncrementalTrackingSolver(Solver):
    """A solver that keeps track of the asserted formulae

    This class provides tracking of the assertions that are stored
    inside the solver, the last executed command and the last solving
    result.

    It requires the extending class to implement the following proxy
    methods:

    * _reset_assertions
    * _add_assertion
    * _solve
    * _push
    * _pop

    The semantics of each function is the same as the non-proxy
    version except for _add_assertion that is supposed to return a
    result (of any type) that will constitute the elements of the
    self.assertions list.
    """

    def __init__(self, environment, logic, **options):
        """See py:func:`Solver.__init__()`."""
        Solver.__init__(self, environment, logic, **options)

        self._last_result = None
        self._last_command = None

        self._assertion_stack = []
        self._backtrack_points = []

    @property
    def last_command(self):
        """Returns the name of the last executed command"""
        return self._last_command

    @property
    def last_result(self):
        """Returns the result of the last call to solve().

        Returns True, False or "unknown": the last result of the last
        call to solve(). If solve has never been called, None is
        returned
        """
        return self._last_result

    @property
    @clear_pending_pop
    def assertions(self):
        """Returns the list of assertions that are still in the solver.

        Returns the list of results of calls to _add_assertion() that
        are still asserted in the solver
        """
        return self._assertion_stack

    def _reset_assertions(self):
        raise NotImplementedError

    def reset_assertions(self):
        self._reset_assertions()
        self._assertion_stack = []
        self._last_command = "reset_assertions"

    def _add_assertion(self, formula, named=None):
        """Assert the formula in the solver.

        This must return the asserted formula (as an FNode) exactly as
        it was asserted in the solver, thus accounting for rewritings,
        simplifications, etc.

        :returns: The asserted formula
        :rtype: :py:class:
        """
        raise NotImplementedError

    def add_assertion(self, formula, named=None):
        tracked = self._add_assertion(formula, named=named)
        self._assertion_stack.append(tracked)
        self._last_command = "assert"

    def _solve(self, assumptions=None):
        raise NotImplementedError

    def solve(self, assumptions=None):
        try:
            res = self._solve(assumptions=assumptions)
            self._last_result = res
            return res
        except SolverReturnedUnknownResultError:
            self._last_result = "unknown"
            raise

        finally:
            self._last_command = "solve"

    def _push(self, levels=1):
        raise NotImplementedError

    def push(self, levels=1):
        self._push(levels=levels)
        point = len(self._assertion_stack)
        for _ in range(levels):
            self._backtrack_points.append(point)
        self._last_command = "push"

    def _pop(self, levels=1):
        raise NotImplementedError

    def pop(self, levels=1):
        self._pop(levels=levels)
        for _ in range(levels):
            point = self._backtrack_points.pop()
            self._assertion_stack = self._assertion_stack[0:point]
        self._last_command = "pop"


class UnsatCoreSolver(object):
    """A solver supporting unsat core extraction"""

    UNSAT_CORE_SUPPORT = True

    def _check_unsat_core_config(self):
        if self.options.unsat_cores_mode is None:
            raise SolverNotConfiguredForUnsatCoresError

        if self.last_result is None or self.last_result:
            raise SolverStatusError("The last call to solve() was not" \
                                    " unsatisfiable")

        if self.last_command != "solve":
            raise SolverStatusError("The solver status has been modified by a" \
                                    " '%s' command after the last call to" \
                                    " solve()" % self.last_command)

    def get_unsat_core(self):
        """Returns the unsat core as a set of formulae.

        After a call to solve() yielding UNSAT, returns the unsat core
        as a set of formulae
        """
        raise NotImplementedError


    def get_named_unsat_core(self):
        """Returns the unsat core as a dict of names to formulae.

        After a call to solve() yielding UNSAT, returns the unsat core as a
        dict of names to formulae
        """
        raise NotImplementedError


class Model(object):
    """An abstract Model for a Solver.

    This class provides basic services to operate on a model returned
    by a solver. This class is used as superclass for more specific
    Models, that are solver dependent or by the EagerModel class.
    """

    def __init__(self, environment):
        self.environment = environment
        self._converter = None

    def get_value(self, formula, model_completion=True):
        """Returns the value of formula in the current model (if one exists).

        If model_completion is True, then variables not appearing in the
        assignment are given a default value, otherwise an error is generated.

        This is a simplified version of the SMT-LIB function get_values .
        """
        raise NotImplementedError

    def get_values(self, formulae, model_completion=True):
        """Evaluates the values of the formulae in the current model.

        Evaluates the values of the formulae in the current model
        returning a dictionary.
        """
        res = {}
        for f in formulae:
            v = self.get_value(f, model_completion=model_completion)
            res[f] = v
        return res

    def get_py_value(self, formula, model_completion=True):
        """Returns the value of formula as a python type.

        E.g., Bool(True) is translated into True.
        This simplifies writing code that branches on values in the model.
        """
        res = self.get_value(formula, model_completion=model_completion)
        assert res.is_constant()
        return res.constant_value()

    def get_py_values(self, formulae, model_completion=True):
        """Returns the values of the formulae as python types.

        Returns the values of the formulae as python types. in the
        current model returning a dictionary.
        """
        res = {}
        for f in formulae:
            v = self.get_py_value(f, model_completion=model_completion)
            res[f] = v
        return res

    def satisfies(self, formula, solver=None):
        """Checks whether the model satisfies the formula.

        The optional solver argument is used to complete partial
        models.
        """

        subs = self.get_values(formula.get_free_variables())
        simp = formula.substitute(subs).simplify()
        if simp.is_true():
            return True
        if simp.is_false():
            return False

        free_vars = simp.get_free_variables()
        if  len(free_vars) > 0:
            # Partial model
            return False

        if self.environment.enable_div_by_0 and solver is not None:
            # Models might not be simplified to a constant value
            # if there is a division by zero. We find the
            # division(s) and ask the solver for a replacement
            # expression.
            stack = [simp]
            div_0 = []
            while stack:
                x = stack.pop()
                if x.is_constant():
                    pass
                elif x.is_div() and x.arg(1).is_zero():
                    div_0.append(x)
                stack += x.args()

            subs = self.get_values(div_0)
            simp = simp.substitute(subs).simplify()
            return simp.is_true()
        return False

    @property
    def converter(self):
        """Get the Converter associated with the Solver."""
        return self._converter

    @converter.setter
    def converter(self, value):
        self._converter = value

    def __getitem__(self, idx):
        return self.get_value(idx, model_completion=True)

    def __str__(self):
        return "\n".join([ "%s := %s" % (var, value) for (var, value) in self])


class Converter(object):
    """A Converter implements functionalities to convert expressions.

    There are two key methods: convert() and back().
    The first performs the forward conversion (pySMT -> Solver API),
    the second performs the backwards conversion (Solver API -> pySMT)
    """

    def convert(self, formula):
        """Convert a PySMT formula into a Solver term."""
        raise NotImplementedError

    def back(self, expr):
        """Convert an expression of the Solver into a PySMT term."""
        raise NotImplementedError



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

def error():
    print("Invalid option try 'install' or 'shell'")
    exit(-1)

if __name__ == "__main__":
    import sys

    cmd = error
    if len(sys.argv) >= 2:
        sys.argv = sys.argv[1:]
        if sys.argv[0] == 'install':
            import pysmt.cmd.install
            cmd = pysmt.cmd.install.main
        elif sys.argv[0] == 'shell':
            import pysmt.cmd.shell
            cmd = pysmt.cmd.shell.main
    cmd()



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import unittest
import tempfile, os

from pysmt.parsing import HRParser, parse
from pysmt.shortcuts import Iff, Symbol, And, Or, LE, Real, Plus, Minus
from pysmt.test.examples import get_example_formulae
from pysmt.exceptions import NoSolverAvailableError
from pysmt.test import TestCase
from pysmt.typing import REAL


class TestHRParser(TestCase):

    def test_precedences(self):
        p = HRParser()

        a,b,c = (Symbol(v) for v in "abc")
        x,y = (Symbol(v, REAL) for v in "xy")

        tests = []
        tests.append(("a | b & c", Or(a, And(b,c))))
        tests.append(("a & b | c", Or(And(a, b) ,c)))

        f1 = LE(Plus(Plus(x, y), Real(5)), Real(7))
        f2 = LE(Plus(x, Real(5)), Minus(Real(7), y))
        tests.append(("x + y + 5.0 <= 7.0", f1))
        tests.append(("x + 5.0 <= 7.0 - y", f2))
        tests.append(("x + y + 5.0 <= 7.0 & x + 5.0 <= 7.0 - y", And(f1, f2)))
        tests.append(("x + 5.0 <= 7.0 - y | x + y + 5.0 <= 7.0 & x + 5.0 <= 7.0 - y",
                      Or(f2, And(f1, f2))))

        for string, formula in tests:
            self.assertEqual(p.parse(string), formula)
            self.assertEqual(parse(string), formula)


    def test_file_parsing(self):
        fname = None
        try:
            fd, fname = tempfile.mkstemp()
            os.close(fd)

            with open(fname, "w") as f:
                f.write("a | b & c\n")

            a,b,c = (Symbol(v) for v in "abc")
            p = HRParser()

            expected = Or(a, And(b,c))
            parsed = p.parse_fname(fname)

            self.assertEqual(parsed, expected)

        finally:
            if fname is not None:
                os.unlink(fname)

    def test_examples(self):
        p = HRParser()
        for (f, _, _, _) in get_example_formulae():
            s = f.serialize()
            res = p.parse(s)
            check = (res == f)
            if not check:
                try:
                    self.assertValid(Iff(res, f))
                except NoSolverAvailableError:
                    pass



if __name__ == '__main__':
    unittest.main()



# Perform ALL-SMT on Theory atoms
#
# This example shows:
#  - How to get the logic of a formula
#  - How to enumerate models
#  - How to extract a partial model
#  - How to use the special operator EqualsOrIff
#
from pysmt.shortcuts import Solver, Not, And, Symbol, Or
from pysmt.shortcuts import LE, GE, Int, Plus, Equals, EqualsOrIff
from pysmt.typing import INT
from pysmt.oracles import get_logic

def all_smt(formula, keys):
    target_logic = get_logic(formula)
    print("Target Logic: %s" % target_logic)
    with Solver(logic=target_logic) as solver:
        solver.add_assertion(formula)
        while solver.solve():
            partial_model = [EqualsOrIff(k, solver.get_value(k)) for k in keys]
            print(partial_model)
            solver.add_assertion(Not(And(partial_model)))


A0 = Symbol("A0", INT)
A1 = Symbol("A1", INT)
A2 = Symbol("A2", INT)

f = And(GE(A0, Int(0)), LE(A0, Int(5)),
        GE(A1, Int(0)), LE(A1, Int(5)),
        GE(A2, Int(0)), LE(A2, Int(5)),
        Equals(Plus(A0, A1, A2), Int(8)))

all_smt(f, [A0, A1, A2])

# By using the operator EqualsOrIff, we can mix theory and bool variables
x = Symbol("x")
y = Symbol("y")
f = And(f, Or(x,y))

all_smt(f, [A0, A1, A2, x])



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from pysmt.shortcuts import *
from pysmt.typing import INT, REAL, FunctionType, BV16
from pysmt.logics import UFLRA, UFLIRA
from pysmt.test import TestCase, main
from pysmt.test import skipIfNoSolverForLogic
from pysmt.exceptions import PysmtModeError, PysmtValueError


class TestEUF(TestCase):

    @skipIfNoSolverForLogic(UFLIRA)
    def test_euf(self):
        ftype1 = FunctionType(REAL, [REAL])
        ftype2 = FunctionType(REAL, [REAL, INT])

        f = Symbol("f", ftype1)
        g = Symbol("g", ftype2)

        check = Equals(Function(f, [Real(1)]), Function(g, (Real(2), Int(4))))

        self.assertSat(check, logic=UFLIRA,
                       msg="Formula was expected to be sat")


    def test_infix(self):
        ftype1 = FunctionType(REAL, [REAL])
        ftype2 = FunctionType(REAL, [REAL, INT])
        f = Symbol("f", ftype1)
        g = Symbol("g", ftype2)

        with self.assertRaises(PysmtModeError):
            f(1.0)

        get_env().enable_infix_notation = True

        infix = Equals(f(1.0), g(2.0, 4))
        explicit = Equals(Function(f, [Real(1.0)]), Function(g, [Real(2.0), Int(4)]))
        self.assertEqual(infix, explicit)

        ftype1 = FunctionType(REAL, [BV16])
        ftype2 = FunctionType(BV16, [INT, BV16])
        f = Symbol("bvf", ftype1)
        g = Symbol("bvg", ftype2)
        infix = Equals(f(g(2, 6)), Real(0))
        explicit = Equals(Function(f, [Function(g, [Int(2), BV(6, 16)])]), Real(0))
        self.assertEqual(infix, explicit)

        with self.assertRaises(PysmtValueError):
            f(BV(6, 16), BV(8, 16))

        ftype3 = FunctionType(REAL, [])
        h = Symbol("h", ftype3)
        with self.assertRaises(PysmtValueError):
            h()

    @skipIfNoSolverForLogic(UFLRA)
    def test_quantified_euf(self):
        ftype1 = FunctionType(REAL, [REAL, REAL])

        plus = Symbol("plus", ftype1)
        x = Symbol('x', REAL)
        y = Symbol('y', REAL)
        z = Symbol('z', REAL)

        axiom = ForAll([x, y], Equals(Function(plus, (x, y)), Plus(x, y)))

        test1 = Equals(Plus(z, Real(4)), Function(plus, (Real(4), z)))
        test2 = Equals(Function(plus, (Real(5), Real(4))), Real(9))

        check = Implies(axiom, And(test1, test2))

        self.assertValid(check, logic=UFLRA,
                        msg="Formula was expected to be valid")


    def test_simplify(self):
        ftype1 = FunctionType(REAL, [REAL, REAL])

        plus = Symbol("plus", ftype1)
        x = Symbol('x', REAL)
        y = Symbol('y', REAL)
        z = Symbol('z', REAL)

        f = Function(plus, (Minus(Real(5), Real(5)),
                            Plus(y, Minus(z, z))))
        self.assertEqual(Function(plus, (Real(0), y)),
                          f.simplify())

if __name__ == '__main__':
    main()



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import pysmt.logics
from pysmt.shortcuts import And, Symbol, Exists, FALSE, ForAll, Or, TRUE
from pysmt.shortcuts import qelim
from pysmt.exceptions import InternalSolverError, NoSolverAvailableError
from pysmt.test import TestCase, main
from pysmt.test.examples import get_example_formulae


class TestShannon(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        self.x, self.y = Symbol("x"), Symbol("y")

    def test_exists(self):
        f = Exists([self.x], And(self.x, self.y))
        g = qelim(f, solver_name="shannon")
        g = g.simplify()
        self.assertEqual(g, self.y)

    def test_forall(self):
        f = ForAll([self.x], And(self.x, self.y))
        g = qelim(f, solver_name="shannon")
        g = g.simplify()
        self.assertEqual(g, FALSE())

    def test_multiple(self):
        f = ForAll([self.x, self.y], Or(self.x, self.y))
        g = qelim(f, solver_name="shannon")
        g = g.simplify()
        self.assertEqual(g, FALSE())

    def test_nested(self):
        f = Exists([self.x], ForAll([self.y], Or(self.x, self.y)))
        g = qelim(f, solver_name="shannon")
        g = g.simplify()
        self.assertEqual(g, TRUE())

    def test_examples_solving(self):
        for example in get_example_formulae():
            if example.logic != pysmt.logics.BOOL:
                continue

            fv = example.expr.get_free_variables()
            f = Exists(fv, example.expr)
            g = qelim(f, solver_name="shannon").simplify()
            if example.is_sat:
                self.assertTrue(g.is_true())
            else:
                self.assertTrue(g.is_false())

            f = ForAll(fv, example.expr)
            g = qelim(f, solver_name="shannon").simplify()
            if example.is_valid:
                self.assertTrue(g.is_true())
            else:
                self.assertTrue(g.is_false())

    def test_w_theory(self):
        for example in get_example_formulae():
            f = example.expr
            if example.logic.quantifier_free: continue
            try:
                res = qelim(f, solver_name="shannon")
                self.assertIsNotNone(res, f)
            except NoSolverAvailableError:
                self.assertTrue(example.logic > pysmt.logics.BOOL, example)
            except InternalSolverError:
                self.assertTrue(example.logic > pysmt.logics.BOOL, example)


if __name__ == '__main__':
    main()



# Copyright 2014 Andrea Micheli and Marco Gario
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import os
import glob

from pysmt.cmd.installers.base import SolverInstaller


class Z3Installer(SolverInstaller):

    SOLVER = "z3"

    def __init__(self, install_dir, bindings_dir, solver_version,
                 mirror_link=None):

        SolverInstaller.__init__(self, install_dir=install_dir,
                                 bindings_dir=bindings_dir,
                                 solver_version=solver_version,
                                 mirror_link=mirror_link)

    def compile(self):
        SolverInstaller.run_python("-m pip install --upgrade --target=%s z3-solver==%s" % (self.bindings_dir, self.solver_version))

    def get_installed_version(self):
        return self.get_installed_version_script(self.bindings_dir, "z3")



# Perform Quantifier Elimination of an LRA formula
#
# This example requires MathSAT or Z3 to be installed
#
# In Quantifier Elimination (QE) we take a formula with quantifiers,
# and rewrite it in an equivalent formula without quantifiers
#
# Example:
#  forall x. exists y. phi(x,y) = exists y . phi'(y)
#
#  If x is a Boolean variable, then we can apply Shannon's expansion
#  to perform the quantifier elimination
#
#  phi'(y) = exists y. phi(FALSE, y) /\ phi(TRUE, y)
#
#  The semantics of forall tells us that the expression must be
#  satisfiable for both values of x: x=TRUE and x=FALSE.
#
# If x is a Rational variable, we cannot enumerate all its possible
# values. One technique to deal with quantifier elimination in LRA is
# called Fourier-Motzkin [1]
#
# [1] https://en.wikipedia.org/wiki/Fourier%E2%80%93Motzkin_elimination
from pysmt.shortcuts import Symbol, ForAll, Exists, qelim
from pysmt.typing import REAL

# (x,y) is a point in a 2d space.

x = Symbol("x", REAL)
y = Symbol("y", REAL)

# phi(x,y) is true if (x,y) is within a given area. In particular, we
# pick a rectangular area of size 5x10: the bottom-left corner has
# coordinate (0,0), the top-right has coordinate (5, 10).
#

rect = (x >= 0.0) & (x <= 5.0) & \
      (y >= 0.0) & (y <= 10.0)

# The first expression that we build asks if for any value of x we can
# define a value of y that would satisfy the expression above.
f1 = ForAll([x], Exists([y], rect))

# This is false, because we know that if we pick x=11, the formula
# above cannot be satisfied. Eliminating all the symbols in an
# expression is the same as solving the expression.
#
# The function to perform quantifier elimination is called qelim:
qf_f1 = qelim(f1)
print(qf_f1)

# We can restrict the values that x can adopt by imposing them as
# precondition.
#
# If we perform quantifier elimination on this expression we obtain an
# unsurprising result:
#
f2 = ForAll([x], Exists([y], ((x >= 0.0) & (x <= 4.0)).Implies(rect)))
qf_f2 = qelim(f2)
print(qf_f2)

# The power of quantifier elimination lies in the possibility of
# representing sets of solutions. Lets introduce a new symbol z, that
# represents the grid distance (aka Manhattan distance) of (x,y) from
# the origin.
z = Symbol("z", REAL)
distance= z.Equals(x + y)

# We want to characterize the set of points in the rectangle, in terms
# of their grid distance from the origin. In other words, we want to
# constraint z to be able to assume values that are possible within
# the rectangle. We want a formula in z that is satisfiable iff there
# is a value for z s.t.  exists a value for x,y within the rectangle.
# An example of the type of information that we expect to see is the
# maximum and minimum value of z.
f3 = Exists([x,y], rect & distance)
qe_f3 = qelim(f3)
print(qe_f3.serialize())

# Depending on the solver in use you might get a different
# result. Try: qelim(f3, solver_name="msat_fm")
#
# MathSAT Fourier Motzkin (FM) returns a compact result: (0.0
# <= z) & (z <= 15.0) , meaning that the further point in grid
# distance is at most 15.0 units away from the origin.
#
# By definition, the expression obtained after performing quantifier
# elimination is in the quantifier free fragment of the logic.  It is
# therefore possible to remove the quantifiers using qelim, and use a
# solver that does not support quantifiers natively to solve the new
# formula.
from pysmt.oracles import get_logic
print(get_logic(f3), get_logic(qe_f3))



# Copyright 2014 Andrea Micheli and Marco Gario
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from pysmt.cmd.installers.base import SolverInstaller


class CuddInstaller(SolverInstaller):

    SOLVER = "bdd"

    def __init__(self, install_dir, bindings_dir, solver_version,
                 mirror_link=None, git_version='HEAD'):
        archive_name = "repycudd-%s.tar.gz" % git_version
        native_link = "https://codeload.github.com/pysmt/repycudd/tar.gz/%s" % git_version
        SolverInstaller.__init__(self, install_dir=install_dir,
                                 bindings_dir=bindings_dir,
                                 solver_version=solver_version,
                                 archive_name=archive_name,
                                 native_link=native_link,
                                 mirror_link=mirror_link)
        self.git_version = git_version


    def compile(self):
        # Select the correct Makefile to be used
        makefile = "Makefile"
        if self.architecture == "x86_64":
            makefile = "Makefile_64bit"

        swig = "swig"
        swig_version = SolverInstaller.run("swig -version", get_output=True)
        if '4.0.1' in swig_version or '4.0.0' in swig_version:
            print("WARNING: the BDD solver does not work with Swig4 < 4.0.2. Fallback to Swig3")
            swig = "swig3.0" # This is the Ubuntu naming of the executable

        import sysconfig
        PYTHON_INCLUDE_DIR = sysconfig.get_path("include")
        SolverInstaller.run("make -C %s -f %s PYTHON_INCL=-I%s SWIG=%s" %
                            (self.extract_path, makefile, PYTHON_INCLUDE_DIR, swig))


    def move(self):
        SolverInstaller.mv(os.path.join(self.extract_path, "repycudd.py"),
                           self.bindings_dir)
        SolverInstaller.mv(os.path.join(self.extract_path, "_repycudd.so"),
                           self.bindings_dir)


    def get_installed_version(self):
        return self.get_installed_version_script(self.bindings_dir, "cudd")



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""Provides the most used functions in a nicely wrapped API.

This module defines a global environment, so that most methods can be
called without the need to specify an environment or a FormulaManager.
Functions trying to access the global environment should use the
method get_env(). Keep in mind that the global state of the
environment might lead to inconsistency and unexpected bugs. This is
particularly true for tests. For tests it is recommended to perform an
environment reset in the setUp phase, to be guaranteed that a fresh
environment is used (this is the default behavior of
:py:class:`pysmt.test.TestCase` ).
"""

# Enable default deprecation warnings!
import warnings
warnings.filterwarnings('default', module='pysmt')

import pysmt.configuration as config
import pysmt.environment
import pysmt.typing as types
import pysmt.smtlib.parser
import pysmt.smtlib.script
import pysmt.smtlib.printers

# Import types from shortcuts
from pysmt.typing import INT, BOOL, REAL, BVType, FunctionType, ArrayType, Type
assert INT or BOOL or REAL or BVType or FunctionType or ArrayType or Type


def get_env():
    """Returns the global environment.

    :returns: The global environment
    :rtype: Environment
    """
    return pysmt.environment.get_env()


def reset_env():
    """Resets the global environment, and returns the new one.

    :returns: A new environment after resetting the global environment
    :rtype: Environment
    """
    return pysmt.environment.reset_env()


# Enable by default infix notation
get_env().enable_infix_notation = True


##### Shortcuts for FormulaManager #####
def get_type(formula):
    """Returns the type of the formula.

    :param formula: The target formula
    :type  formula: FNode
    :returns: The type of the formula
    """
    return get_env().stc.get_type(formula)


def simplify(formula):
    """Returns the simplified version of the formula.

    :param formula: The target formula
    :type  formula: FNode
    :returns: The simplified version of the formula
    :rtype: Fnode
    """
    return get_env().simplifier.simplify(formula)


def substitute(formula, subs):
    """Applies the substitutions defined in the dictionary to the formula.

    :param formula: The target formula
    :type  formula: FNode
    :param subs: Specify the substitutions to apply to the formula
    :type  subs: A dictionary from FNode to FNode
    :returns: Formula after applying the substitutions
    :rtype: Fnode
    """
    return get_env().substituter.substitute(formula, subs)


def serialize(formula, threshold=None):
    """Provides a string representing the formula.

    :param formula: The target formula
    :type  formula: FNode
    :param threshold: Specify the threshold
    :type  formula: Integer
    :returns: A string representing the formula
    :rtype: string
    """
    return get_env().serializer.serialize(formula,
                                          threshold=threshold)

def get_free_variables(formula):
    """Returns the free variables of the formula.

    :param formula: The target formula
    :type  formula: FNode
    :returns: Free variables in the formula
    """
    return get_env().fvo.get_free_variables(formula)


def get_atoms(formula):
    """Returns the set of atoms of the formula.

    :param formula: The target formula
    :type  formula: FNode
    :returns: the set of atoms of the formula
    """
    return get_env().ao.get_atoms(formula)


def get_formula_size(formula, measure=None):
    """Returns the size of the formula as measured by the given counting type.

    See pysmt.oracles.SizeOracle for details.

    :param formula: The target formula
    :type  formula: FNode
    :param measure: Specify the measure/counting type
    :returns: The size of the formula as measured by the given counting type.
    """
    return get_env().sizeo.get_size(formula, measure)


##### Nodes Creation #####

def ForAll(variables, formula):
    r""".. math:: \forall v_1, \cdots, v_n . \varphi(v_1, \cdots, v_n)"""
    return get_env().formula_manager.ForAll(variables, formula)


def Exists(variables, formula):
    r""".. math:: \exists v_1, \cdots, v_n . \varphi(v_1, \cdots, v_n)"""
    return get_env().formula_manager.Exists(variables, formula)


def Function(vname, params):
    r""".. math:: vname(p_1, \cdots, p_n)"""
    return get_env().formula_manager.Function(vname, params)


def Not(formula):
    r""".. math:: \lnot \varphi"""
    return get_env().formula_manager.Not(formula)


def Implies(left, right):
    r""".. math:: l \rightarrow r"""
    return get_env().formula_manager.Implies(left, right)


def Iff(left, right):
    r""".. math:: l \leftrightarrow r """
    return get_env().formula_manager.Iff(left, right)


def GE(left, right):
    r""".. math:: l \ge r"""
    return get_env().formula_manager.GE(left, right)


def Minus(left, right):
    r""".. math:: l - r """
    return get_env().formula_manager.Minus(left, right)


def Times(*args):
    r""".. math:: x_1 \times x_2 \cdots \times x_n"""
    return get_env().formula_manager.Times(*args)


def Pow(left, right):
    r""".. math:: l ^ r"""
    return get_env().formula_manager.Pow(left, right)


def Div(left, right):
    r""".. math:: \frac{l}{r}"""
    return get_env().formula_manager.Div(left, right)


def Equals(left, right):
    r""".. math:: l = r"""
    return get_env().formula_manager.Equals(left, right)


def NotEquals(left, right):
    r""".. math:: l != r"""
    return get_env().formula_manager.NotEquals(left, right)

def GT(left, right):
    r""".. math:: l > r"""
    return get_env().formula_manager.GT(left, right)


def LE(left, right):
    r""".. math:: l \le r"""
    return get_env().formula_manager.LE(left, right)


def LT(left, right):
    r""".. math:: l < r"""
    return get_env().formula_manager.LT(left, right)


def Ite(iff, left, right):
    r""".. math:: \text{ If } i \text{ Then } l \text{ Else } r"""
    return get_env().formula_manager.Ite(iff, left, right)


def Symbol(name, typename=types.BOOL):
    """Returns a symbol with the given name and type.

    :param name: Specify the name
    :param typename: Specify the typename
    :returns: A symbol with the given name and type
    """
    return get_env().formula_manager.Symbol(name, typename)


def FreshSymbol(typename=types.BOOL, template=None):
    """Returns a symbol with a fresh name and given type.

    :param typename: Specify the typename
    :param template: Specify the template
    :returns: A symbol with a fresh name and a given type
    """
    return get_env().formula_manager.FreshSymbol(typename, template)


def Int(value):
    """Returns an Integer constant with the given value.

    :param value: Specify the value
    :returns: An Integer constant with the given value
    """
    return get_env().formula_manager.Int(value)


def Bool(value):
    """Returns a Boolean constant with the given value.

    :param value: Specify the value
    :returns: A Boolean constant with the given value
    """
    return get_env().formula_manager.Bool(value)


def Real(value):
    """Returns a Real constant with the given value.

    :param value: Specify the value
    :returns: A Real constant with the given value
    """
    return get_env().formula_manager.Real(value)


def String(value):
    """Returns a String constant with the given value."""
    return get_env().formula_manager.String(value)


def TRUE():
    """Returns the Boolean constant TRUE.

    returns: The Boolean constant TRUE
    """
    return get_env().formula_manager.TRUE()


def FALSE():
    """Returns the Boolean constant FALSE.

    returns: The Boolean constant FALSE
    """
    return get_env().formula_manager.FALSE()


def And(*args):
    r""".. math:: \varphi_0 \land \cdots \land \varphi_n """
    return get_env().formula_manager.And(*args)


def Or(*args):
    r""".. math:: \varphi_0 \lor \cdots \lor \varphi_n """
    return get_env().formula_manager.Or(*args)


def Plus(*args):
    r""".. math:: \varphi_0 + \cdots + \varphi_n """
    return get_env().formula_manager.Plus(*args)


def ToReal(formula):
    """Explicit cast of a term into a Real term."""
    return get_env().formula_manager.ToReal(formula)


def AtMostOne(*args):
    """At most one can be true at anytime.

    Cardinality constraint over a set of boolean expressions.
    """
    return get_env().formula_manager.AtMostOne(*args)


def ExactlyOne(*args):
    """Given a set of boolean expressions requires that exactly one holds."""
    return get_env().formula_manager.ExactlyOne(*args)


def AllDifferent(*args):
    """Given a set of non-boolean expressions, requires that each of them
    has value different from all the others
    """
    return get_env().formula_manager.AllDifferent(*args)


def Xor(left, right):
    """Returns the XOR of left and right

    :param left: Specify the left BV
    :type  left: FNode
    :param right: Specify the right BV
    :type  right: FNode
    :returns: The XOR of left and right
    """
    return get_env().formula_manager.Xor(left, right)


def Min(*args):
    """Minimum over a set of real or integer terms."""
    return get_env().formula_manager.Min(*args)


def Max(*args):
    """Maximum over a set of real or integer terms"""
    return get_env().formula_manager.Max(*args)


def EqualsOrIff(left, right):
    """Returns Equals() or Iff() depending on the type of the arguments.

    This can be used to deal with ambiguous cases where we might be
    dealing with both Theory and Boolean atoms.
    """
    return get_env().formula_manager.EqualsOrIff(left, right)


#
# Bit Vectors
#
def BV(value, width=None):
    """Returns a constant of type BitVector.

    value can be either:
    - a string of 0s and 1s
    - a string starting with "#b" followed by a sequence of 0s and 1s
    - an integer number s.t. 0 <= value < 2**width

    In order to create the BV representation of a signed integer,
    the SBV() method shall be used.

    :param value: Specify the value
    :param width: Specify the width
    :returns: A constant of type BitVector
    :rtype: FNode
    """
    return get_env().formula_manager.BV(value, width)


def SBV(value, width=None):
    """Returns a constant of type BitVector interpreting the sign.

    If the specified value is an integer, it is converted in the
    2-complement representation of the given number, otherwise the
    behavior is the same as BV().

    :param value: Specify the value
    :param width: Specify the width of the BV
    :returns: A constant of type BitVector interpreting the sign.
    :rtype: FNode
    """
    return get_env().formula_manager.SBV(value, width)


def BVOne(width=None):
    """Returns the unsigned one constant BitVector.

    :param width: Specify the width of the BitVector
    :returns: The unsigned one constant BitVector
    :rtype: FNode
    """
    return get_env().formula_manager.BVOne(width)


def BVZero(width=None):
    """Returns the zero constant BitVector.

    :param width: Specify the width of the BitVector
    :returns: The unsigned zero constant BitVector
    :rtype: FNode
    """
    return get_env().formula_manager.BVZero(width)


def BVNot(formula):
    """Returns the bitwise negation of the bitvector

    :param formula: The target formula
    :returns: The bitvector Not(bv)
    :rtype: FNode
    """
    return get_env().formula_manager.BVNot(formula)


def BVAnd(*args):
    """Returns the Bit-wise AND of bitvectors of the same size.
    If more than 2 arguments are passed, a left-associative formula is generated.

    :param *args: Specify the bitvectors
    :returns: The bit-wise AND of the bitvectors using left association
    :rtype: FNode
    """
    return get_env().formula_manager.BVAnd(*args)


def BVOr(*args):
    """Returns the Bit-wise OR of bitvectors of the same size.
    If more than 2 arguments are passed, a left-associative formula is generated.

    :param *args: Specify the bitvectors
    :returns: The bit-wise OR of the bitvectors using left association
    :rtype: FNode
    """
    return get_env().formula_manager.BVOr(*args)


def BVXor(left, right):
    """Returns the Bit-wise XOR of two bitvectors of the same size.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The bit-wise XOR of left and right
    :rtype: FNode
    """
    return get_env().formula_manager.BVXor(left, right)


def BVConcat(*args):
    """Returns the Concatenation of the two BVs

    :param args: Specify the bitvectors to concatenate
    :returns: The concatenation of the given BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVConcat(*args)


def BVExtract(formula, start=0, end=None):
    """Returns the slice of formula from start to end (inclusive).

    :param formula: The target formula
    :param start: Specify the start index
    :param end: Specify the end index
    :returns: The slice of formula from start to end (inclusive)
    :rtype: Fnode
    """
    return get_env().formula_manager.BVExtract(formula, start=start, end=end)


def BVULT(left, right):
    """Returns the Unsigned Less-Than comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned Less-Than comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVULT(left, right)


def BVUGT(left, right):
    """Returns the Unsigned Greater-Than comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned Greater-Than comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVUGT(left, right)


def BVULE(left, right):
    """Returns the Unsigned Less-Equal comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned Less-Equal comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVULE(left, right)


def BVUGE(left, right):
    """Returns the Unsigned Greater-Equal comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned Greater-Equal comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVUGE(left, right)


def BVNeg(formula):
    """Returns the arithmetic negation of the BV.

    :param formula: The target formula
    :returns: The arithmetic negation of the formula
    :rtype: FNode
    """
    return get_env().formula_manager.BVNeg(formula)

def BVAdd(*args):
    """Returns the sum of BV.
    If more than 2 arguments are passed, a left-associative formula is generated.

    :param *args: Specify the bitvectors
    :returns: The sum of the bitvectors using left association.
    :rtype: FNode
    """
    return get_env().formula_manager.BVAdd(*args)


def BVSub(left, right):
    """Returns the difference of two BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The difference of the two BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVSub(left, right)


def BVMul(*args):
    """Returns the product of BV.
    If more than 2 arguments are passed, a left-associative formula is generated.

    :param *args: Specify the bitvectors
    :returns: The product of the the bitvectors using left association
    :rtype: FNode
    """
    return get_env().formula_manager.BVMul(*args)


def BVUDiv(left, right):
    """Returns the Unsigned division of the two BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned division of the two BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVUDiv(left, right)


def BVURem(left, right):
    """Returns the Unsigned remainder of the two BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Unsigned remainder of the two BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVURem(left, right)


def BVLShl(left, right):
    """Returns the logical left shift the BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The logical left shift the BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVLShl(left, right)


def BVLShr(left, right):
    """Returns the logical right shift the BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The logical right shift the BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVLShr(left, right)


def BVAShr(left, right):
    """Returns the RIGHT arithmetic shift of the left BV by the number
        of steps specified by the right BV.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The RIGHT arithmetic shift of the left BV by the number
              of steps specified by the right BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVAShr(left, right)


def BVRol(formula, steps):
    """Returns the LEFT rotation of the BV by the number of steps.

    :param formula: The target formula
    :param steps: Specify the number of steps.
    :returns: The LEFT rotation of the BV by the number of steps
    :rtype: FNode
    """
    return get_env().formula_manager.BVRol(formula, steps)


def BVRor(formula, steps):
    """Returns the RIGHT rotation of the BV by the number of steps.

    :param formula: The target formula
    :param steps: Specify the number of steps.
    :returns: The RIGHT rotation of the BV by the number of steps
    :rtype: FNode
    """
    return get_env().formula_manager.BVRor(formula, steps)


def BVZExt(formula, increase):
    """Returns the zero-extension of the BV.

    New bits are set to zero.

    :param formula: The target formula
    :param increase: Specify the increase
    :returns: The extension of the BV
    :rtype: FNode
    """
    return get_env().formula_manager.BVZExt(formula, increase)


def BVSExt(formula, increase):
    """Returns the signed-extension of the BV.

    New bits are set according to the most-significant-bit.

    :param formula: The target formula
    :param increase: Specify the 'increase' value
    :returns: The signed-extension of the BV.
    :rtype: FNode
    """
    return get_env().formula_manager.BVSExt(formula, increase)


def BVSLT(left, right):
    """Returns the Signed Less-Than comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Signed Less-Than comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVSLT(left, right)


def BVSLE(left, right):
    """Returns the Signed Less-Equal comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Signed Less-Than-Equal comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVSLE(left, right)


def BVSGT(left, right):
    """Returns the Signed Greater-Than comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Signed Greater-Than comparison of the two BVs
    :rtype: FNode
    """
    return get_env().formula_manager.BVSGT(left, right)


def BVSGE(left, right):
    """Returns the Signed Greater-Equal comparison of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Signed Greater-Equal comparison of the two BVs.
    :rtype: FNode
    """
    return get_env().formula_manager.BVSGE(left, right)


def BVSDiv(left, right):
    """Returns the Signed division of the two BVs.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The the Signed division of left by right
    :rtype: FNode
    """
    return get_env().formula_manager.BVSDiv(left, right)


def BVSRem(left, right):
    """Returns the Signed remainder of the two BVs

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: The Signed remainder of left divided by right
    :rtype: FNode
    """
    return get_env().formula_manager.BVSRem(left, right)


def BVComp(left, right):
    """Returns a BV of size 1 equal to 0 if left is equal to right,
        otherwise equal to 1.

    :param left: Specify the left bitvector
    :param right: Specify the right bitvector
    :returns: A BV of size 1 equal to 0 if left is equal to right,
              otherwise 1
    :rtype: FNode
    """
    return get_env().formula_manager.BVComp(left, right)

def BVToNatural(formula):
    """Returns the Natural number represented by the BitVector.

    :param formula: Bitvector to be converted
    :returns: An integer between 0 and 2^m-1
    :rtype: FNode
    """
    return get_env().formula_manager.BVToNatural(formula)

# Strings
def StrLength(string):
    """Returns the length of a formula resulting a String"""
    return get_env().formula_manager.StrLength(string)

def StrCharAt(string, index):
    """Returns a single character String at position i.

    s is a string term and i is an integer term. i is the position.
    """
    return get_env().formula_manager.StrCharAt(string, index)

def StrConcat(*args):
    """Returns the concatenation of n Strings.

    s1, s2, ..., and sn are String terms.
    String concatenation takes at least 2 arguments.
    """
    return get_env().formula_manager.StrConcat(*args)

def StrContains(string, target):
    """Returns wether string contains the target.

    s and t are String terms.
    """
    return get_env().formula_manager.StrContains(string, target)

def StrIndexOf(string, target, offset):
    """Returns the position of the first occurrence of target in string after the offset.

    s and t being a non empty strings and i a non-negative integer.
    It returns -1 if the value to search for never occurs.
    """
    return get_env().formula_manager.StrIndexOf(string, target, offset)

def StrReplace(s, t1, t2):
    """Returns a new string where the first occurrence of t1 is replace by t2.

    where s, t1 and t2 are string terms, t1 is non-empty.
    """
    return get_env().formula_manager.StrReplace(s, t1, t2)

def StrSubstr(s, i, j):
    """Returns a substring of s starting at i and ending at j+i.

    where s is a string term and i, j are integer terms.
    """
    return get_env().formula_manager.StrSubstr(s, i, j)

def StrPrefixOf(s, t):
    """Returns whether the s is a prefix of the string t.

    where s and t are string terms.
    """
    return get_env().formula_manager.StrPrefixOf(s, t)

def StrSuffixOf(s, t):
    """Returns whether the string s is a suffix of the string t.

    where s and t are string terms.
    """
    return get_env().formula_manager.StrSuffixOf(s, t)

def StrToInt(x):
    """Returns the corresponding natural number of s is valid;

    If s is not valid, it returns -1.
    """
    return get_env().formula_manager.StrToInt(x)

def IntToStr(x):
    """Returns the corresponding String representing the natural number x.

    where x is an integer term. If x is not a natural number it
    returns the empty String.
    """
    return get_env().formula_manager.IntToStr(x)

#
# Arrays
#
def Select(array, index):
    """Returns a SELECT application on the array at the given index

    :param array: Specify the array
    :param index: Specify the index
    :returns: A SELECT application on array at index
    :rtype: FNode
    """
    return get_env().formula_manager.Select(array, index)


def Store(array, index, value):
    """Returns a STORE application with given value on array at the given index

    :param array: Specify the array
    :param index: Specify the index
    :returns: A STORE on the array at the given index with the given value
    :rtype: FNode
    """
    return get_env().formula_manager.Store(array, index, value)


def Array(idx_type, default, assigned_values=None):
    """Returns an Array with the given index type and initialization.

    If assigned_values is specified, then it must be a map from
    constants of type idx_type to values of the same type as default
    and the array is initialized correspondingly.

    :param idx_type: Specify the index type
    :param default: Specify the default values
    :param assigned_values: Specify the assigned values
    :returns: A node representing an array having index type equal to
              idx_type, initialized with default values. If assigned_values
              is specified, then it must be a map from constants of type
              idx_type to values of the same type as default and the array
              is initialized correspondingly.
    :rtype: FNode
    """
    return get_env().formula_manager.Array(idx_type, default, assigned_values)


##
## Shortcuts for Solvers Factory
##
def Solver(name=None, logic=None, **kwargs):
    """Returns a solver.

    :param name: Specify the name of the solver
    :param logic: Specify the logic that is going to be used.
    :rtype: Solver
    """
    return get_env().factory.Solver(name=name,
                                    logic=logic,
                                    **kwargs)

def UnsatCoreSolver(name=None, logic=None, unsat_cores_mode="all", **kwargs):
    """Returns a solver supporting unsat core extraction.

    :param name: Specify the name of the solver
    :param logic: Specify the logic that is going to be used.
    :param unsat_cores_mode: Specify the unsat cores mode.
    :returns: A solver supporting unsat core extraction.
    :rtype: Solver
    """
    return get_env().factory.UnsatCoreSolver(name=name,
                                             logic=logic,
                                             unsat_cores_mode=unsat_cores_mode,
                                             **kwargs)


def QuantifierEliminator(name=None, logic=None):
    """Returns a quantifier eliminator.

    :param name: Specify the name of the solver
    :param logic: Specify the logic that is going to be used.
    :returns: A quantifier eliminator with the specified
              name and logic
    :rtype: QuantifierEliminator
    """
    return get_env().factory.QuantifierEliminator(name=name, logic=logic)


def Interpolator(name=None, logic=None):
    """Returns an interpolator

    :param name: Specify the name of the solver
    :param logic: Specify the logic that is going to be used.
    :returns: An interpolator
    :rtype: Interpolator
    """
    return get_env().factory.Interpolator(name=name, logic=logic)


def Portfolio(solvers_set, logic, **options):
    """Creates a portfolio using the specified solvers.

    Solver_set is an iterable. Elements of solver_set can be
      1) a name of a solver
      2) a tuple containing a name of a solver and dict of options

    E.g.,
      Portfolio(["msat", "z3"], incremental=True)
    or
      Portfolio([("msat", {"random_seed": 1}), ("msat", {"random_seed": 2})],
               incremental=True)

    Options specified in the Portfolio are shared among all
    solvers, e.g., in the first example all solvers will receive
    the option 'incremental=True'.

    One process will be used for each of the solvers.

    :param solvers_set: Specify set of solvers to be used in the portfolio.
    :param logic: Specify the logic that is going to be used, this
        might restrict the set of solvers in the portfolio.
    :returns: A portfolio solver
    :rtype: Portfolio
    """
    import pysmt.solvers.portfolio as pf
    return pf.Portfolio(solvers_set=solvers_set,
                        logic=logic,
                        environment=get_env(),
                        **options)


def is_sat(formula, solver_name=None, logic=None, portfolio=None):
    """ Returns whether a formula is satisfiable.

    :param formula: The formula to check satisfiability
    :type  formula: FNode
    :param solver_name: Specify the name of the solver to be used
    :type  solver_name: string
    :param logic: Specify the logic that is going to be used
    :param portfolio: A list of solver names to perform portfolio solving.
    :type  portfolio: An iterable of solver names
    :returns: Whether the formula is SAT or UNSAT.
    :rtype: bool
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during is_sat")
        formula = env.formula_manager.normalize(formula)

    return env.factory.is_sat(formula,
                              solver_name=solver_name,
                              logic=logic,
                              portfolio=portfolio)


def get_model(formula, solver_name=None, logic=None):
    """ Similar to :py:func:`is_sat` but returns a model if the formula is
    satisfiable, otherwise None

    :param formula: The target formula
    :param solver_name: Specify the name of the solver
    :param: logic: Specify the logic that is going to be used
    :returns: A model if the formula is satisfiable
    :rtype: Model
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during get_model")
        formula = env.formula_manager.normalize(formula)

    return env.factory.get_model(formula,
                                 solver_name=solver_name,
                                 logic=logic)


def get_implicant(formula, solver_name=None, logic=None):
    """Returns a formula f_i such that Implies(f_i, formula) is valid or None
    if formula is unsatisfiable.

    if complete is set to true, all the variables appearing in the
    formula are forced to appear in f_i.
    :param formula: The target formula
    :param solver_name: Specify the name of the solver
    :param: logic: Specify the logic that is going to be used
    :returns: A formula f_i such that Implies(f_i, formula) is valid or None
              if formula is unsatisfiable.
    :rtype: FNode
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during get_model")
        formula = env.formula_manager.normalize(formula)

    return env.factory.get_implicant(formula,
                                     solver_name=solver_name,
                                     logic=logic)


def get_unsat_core(clauses, solver_name=None, logic=None):
    """Similar to :py:func:`get_model` but returns the unsat core of the
    conjunction of the input clauses

    :param clauses: Specify the list of input clauses
    :param solver_name: Specify the name of the solver_name
    :param logic: Specify the logic that is going to be used
    :returns: The unsat core of the conjunction of the input clauses
    """
    env = get_env()
    clauses = list(clauses)
    if any(c not in env.formula_manager for c in clauses):
        warnings.warn("Warning: Contextualizing formula during get_model")
        clauses = [env.formula_manager.normalize(c) for c in clauses]

    return env.factory.get_unsat_core(clauses,
                                      solver_name=solver_name,
                                      logic=logic)


def is_valid(formula, solver_name=None, logic=None, portfolio=None):
    """Similar to :py:func:`is_sat` but checks validity.

    :param formula: The target formula
    :type  formula: FNode
    :param solver_name: Specify the name of the solver to be used
    :param logic: Specify the logic that is going to be used
    :param portfolio: A list of solver names to perform portfolio solving.
    :returns: Whether the formula is SAT or UNSAT but checks validity
    :rtype: bool
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during is_valid")
        formula = env.formula_manager.normalize(formula)

    return env.factory.is_valid(formula,
                                solver_name=solver_name,
                                logic=logic,
                                portfolio=portfolio)


def is_unsat(formula, solver_name=None, logic=None, portfolio=None):
    """Similar to :py:func:`is_sat` but checks unsatisfiability.

    :param formula: The target formula
    :type  formula: FNode
    :param solver_name: Specify the name of the solver to be used
    :param logic: Specify the logic that is going to be used
    :param portfolio: A list of solver names to perform portfolio solving.
    :returns: Whether the formula is UNSAT or not
    :rtype: bool
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during is_unsat")
        formula = env.formula_manager.normalize(formula)

    return env.factory.is_unsat(formula,
                                solver_name=solver_name,
                                logic=logic,
                                portfolio=portfolio)


def qelim(formula, solver_name=None, logic=None):
    """Performs quantifier elimination of the given formula.

    :param formula: The target formula
    :param solver_name: Specify the name of the solver to be used
    :param logic: Specify the logic that is going to be used
    :returns: A formula after performing quantifier elimination
    :rtype: FNode
    """
    env = get_env()
    if formula not in env.formula_manager:
        warnings.warn("Warning: Contextualizing formula during is_unsat")
        formula = env.formula_manager.normalize(formula)

    return env.factory.qelim(formula,
                             solver_name=solver_name,
                             logic=logic)


def binary_interpolant(formula_a, formula_b, solver_name=None, logic=None):
    """Computes an interpolant of (formula_a, formula_b).

    Returns None if the conjunction is satisfiable

    :param formula_a: Specify formula_a
    :param formula_b: Specify formula_b
    :param solver_name: Specify the name of the solver to be used
    :param logic: Specify the logic that is going to be used
    :returns: An interpolant of (formula_a, formula_b); None
              if the conjunction is satisfiable
    :rtype: FNode or None
    """
    env = get_env()
    formulas = [formula_a, formula_b]
    for i, f in enumerate(formulas):
        if f not in env.formula_manager:
            warnings.warn("Warning: Contextualizing formula during "
                          "binary_interpolant")
            formulas[i] = env.formula_manager.normalize(f)

    return env.factory.binary_interpolant(formulas[0], formulas[1],
                                          solver_name=solver_name,
                                          logic=logic)


def sequence_interpolant(formulas, solver_name=None, logic=None):
    """Computes a sequence interpolant of the formulas.

    Returns None if the conjunction is satisfiable.

    :param formulas: The target formulas
    :param solver_name: Specify the name of the solver to be used
    :param logic: Specify the logic that is going to be used
    :returns: A sequence intepolant of the formulas; None
              if the conjunction is satisfiable
    :rtype: FNode or None
    """
    env = get_env()
    formulas = list(formulas)
    for i, f in enumerate(formulas):
        if f not in env.formula_manager:
            warnings.warn("Warning: Contextualizing formula during "
                          "sequence_interpolant")
            formulas[i] = env.formula_manager.normalize(f)

    return env.factory.sequence_interpolant(formulas,
                                            solver_name=solver_name,
                                            logic=logic)


def read_configuration(config_filename, environment=None):
    """Reads the pysmt configuration of the given file path and applies
    it on the specified environment. If no environment is specified,
    the top-level environment will be used.

    :param config_filename: Specify the name of the config file
    :param environment: Specify the environment
    """
    if environment is None:
        environment = get_env()
    config.configure_environment(config_filename, environment)


def write_configuration(config_filename, environment=None):
    """Dumps the current pysmt configuration to the specified file path

    :param config_filename: Specify the name of the config file
    :param environment: Specify the environment
    """
    if environment is None:
        environment = get_env()
    config.write_environment_configuration(config_filename, environment)


def read_smtlib(fname):
    """Reads the SMT formula from the given file.

    This supports compressed files, if the fname ends in .bz2 .

    :param fname: Specify the filename
    :returns: An SMT formula
    :rtype: FNode
    """
    return pysmt.smtlib.parser.get_formula_fname(fname)


def write_smtlib(formula, fname):
    """Writes the given formula in Smt-Lib format to the given file.

    :param formula: Specify the SMT formula to be written
    :param fname: Specify the filename
    """
    with open(fname, "w") as fout:
        script = pysmt.smtlib.script.smtlibscript_from_formula(formula)
        script.serialize(fout)


def to_smtlib(formula, daggify=True):
    """Returns a Smt-Lib string representation of the formula.

    The daggify parameter can be used to switch from a linear-size
    representation that uses 'let' operators to represent the
    formula as a dag or a simpler (but possibly exponential)
    representation that expands the formula as a tree.

    See :py:class:`SmtPrinter`
    """
    return pysmt.smtlib.printers.to_smtlib(formula, daggify=daggify)



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import math

import pysmt.walkers
from pysmt.walkers import handles
import pysmt.operators as op
import pysmt.typing as types
from pysmt.utils import set_bit
from pysmt.exceptions import PysmtValueError
from pysmt.fnode import FNode


class Simplifier(pysmt.walkers.DagWalker):
    """Perform basic simplifications of the input formula."""

    def __init__(self, env=None):
        pysmt.walkers.DagWalker.__init__(self, env=env)
        self.manager = self.env.formula_manager
        self._validate_simplifications = None
        self.original_walk = self.walk

    @property
    def validate_simplifications(self):
        return self._validate_simplifications

    @validate_simplifications.setter
    def validate_simplifications(self, value):
        """If set to true: checks for equivalence after each simplification.

        NOTE: This can be very expensive, and should be used for debug
        and testing only.
        """
        if value:
            self.walk = self.walk_debug
        else:
            # Restore original walk method
            self.walk = self.original_walk

        self._validate_simplifications = value

    def simplify(self, formula):
        """Performs simplification of the given formula."""
        return self.walk(formula)

    def _get_key(self, formula, **kwargs):
        return formula

    def walk_debug(self, formula, **kwargs):
        from pysmt.shortcuts import Equals, Iff, get_type, is_valid
        from pysmt.typing import BOOL

        if formula in self.memoization:
            return self.memoization[formula]

        args = [self.walk(s, **kwargs) for s in formula.args()]

        f = self.functions[formula.node_type()]
        res = f(formula, args=args, **kwargs)
        ltype = get_type(formula)
        rtype = get_type(res)
        test = Equals(formula, res) if ltype != BOOL else Iff(formula, res)
        assert (ltype == rtype) and is_valid(test,
                                             solver_name=self.validate_simplifications), \
               ("Was: %s \n Obtained: %s\n" % (str(formula), str(res)))
        return res

    def walk_and(self, formula, args, **kwargs):
        if len(args) == 2 and args[0] == args[1]:
            return args[0]

        new_args = set()
        for a in args:
            if a.is_true():
                continue
            if a.is_false():
                return self.manager.FALSE()
            if a.is_and():
                for s in a.args():
                    if self.walk_not(self.manager.Not(s), [s]) in new_args:
                        return self.manager.FALSE()
                    new_args.add(s)
            else:
                if self.walk_not(self.manager.Not(a), [a]) in new_args:
                    return self.manager.FALSE()
                new_args.add(a)

        if len(new_args) == 0:
            return self.manager.TRUE()
        elif len(new_args) == 1:
            return next(iter(new_args))
        else:
            return self.manager.And(new_args)

    def walk_or(self, formula, args, **kwargs):
        if len(args) == 2 and args[0] == args[1]:
            return args[0]

        new_args = set()
        for a in args:
            if a.is_false():
                continue
            if a.is_true():
                return self.manager.TRUE()
            if a.is_or():
                for s in a.args():
                    if self.walk_not(self.manager.Not(s), [s]) in new_args:
                        return self.manager.TRUE()
                    new_args.add(s)
            else:
                if self.walk_not(self.manager.Not(a), [a]) in new_args:
                    return self.manager.TRUE()
                new_args.add(a)

        if len(new_args) == 0:
            return self.manager.FALSE()
        elif len(new_args) == 1:
            return next(iter(new_args))
        else:
            return self.manager.Or(new_args)

    def walk_not(self, formula, args, **kwargs):
        assert len(args) == 1
        args = args[0]
        if args.is_bool_constant():
            l = args.constant_value()
            return self.manager.Bool(not l)
        elif args.is_not():
            return args.arg(0)

        return self.manager.Not(args)

    def walk_iff(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_bool_constant() and sr.is_bool_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l == r)
        elif sl.is_bool_constant():
            if sl.constant_value():
                return sr
            else:
                return self.manager.Not(sr)
        elif sr.is_bool_constant():
            if sr.constant_value():
                return sl
            else:
                return self.manager.Not(sl)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Iff(sl, sr)

    def walk_implies(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_bool_constant():
            l = sl.constant_value()
            if l:
                return sr
            else:
                return self.manager.TRUE()
        elif sr.is_bool_constant():
            r = sr.constant_value()
            if r:
                return self.manager.TRUE()
            else:
                return self.manager.Not(sl)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Implies(sl, sr)

    def walk_equals(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l == r)
        elif sl == sr:
            return self.manager.TRUE()
        else:
            return self.manager.Equals(sl, sr)

    def walk_ite(self, formula, args, **kwargs):
        assert len(args) == 3
        si = args[0]
        st = args[1]
        se = args[2]

        if st == se:
            return st
        elif si.is_bool_constant():
            if si.constant_value():
                return st
            else:
                return se
        else:
            return self.manager.Ite(si, st, se)

    def walk_le(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l <= r)

        # # (le 0 (- X Y)) => (le Y X)
        if sl.is_zero() and sr.is_minus():
            x, y = sr.arg(0), sr.arg(1)
            return self.manager.LE(y, x)
        # (le (- X Y) 0) => (le X Y)
        if sr.is_zero() and sr.is_minus():
            x, y = sr.arg(0), sr.arg(1)
            return self.manager.LE(x, y)
        return  self.manager.LE(sl, sr)

    def walk_lt(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Bool(l < r)
        return self.manager.LT(sl, sr)

    def walk_forall(self, formula, args, **kwargs):
        assert len(args) == 1
        sf = args[0]

        varset = set(formula.quantifier_vars()).intersection(sf.get_free_variables())

        if len(varset) == 0:
            return sf

        return self.manager.ForAll(varset, sf)

    def walk_exists(self, formula, args, **kwargs):
        assert len(args) == 1
        sf = args[0]

        varset = set(formula.quantifier_vars()).intersection(sf.get_free_variables())

        if len(varset) == 0:
            return sf

        return self.manager.Exists(varset, sf)

    def walk_plus(self, formula, args, **kwargs):
        to_sum = []
        to_sub = []
        constant_add = 0
        stack = list(args)
        ttype = self.env.stc.get_type(args[0])
        is_algebraic = False
        while len(stack) > 0:
            x = stack.pop()
            if x.is_constant():
                if x.is_algebraic_constant():
                    is_algebraic = True
                constant_add += x.constant_value()
            elif x.is_plus():
                stack += x.args()
            elif x.is_minus():
                to_sum.append(x.arg(0))
                to_sub.append(x.arg(1))
            elif x.is_times() and x.args()[-1].is_constant():
                const = x.args()[-1]
                const_val = const.constant_value()
                if const_val < 0:
                    new_times = list(x.args()[:-1])
                    if const_val != -1:
                        const_val = -const_val
                        if const.is_algebraic_constant():
                            from pysmt.constants import Numeral
                            const = self.manager._Algebraic(Numeral(const_val))
                        elif ttype.is_real_type():
                            const = self.manager.Real(const_val)
                        else:
                            assert ttype.is_int_type()
                            const = self.manager.Int(const_val)
                        new_times.append(const)
                    new_times = self.manager.Times(new_times)
                    to_sub.append(new_times)
                else:
                    to_sum.append(x)
            else:
                to_sum.append(x)

        const = None
        if is_algebraic:
            from pysmt.constants import Numeral
            const = self.manager._Algebraic(Numeral(constant_add))
        elif ttype.is_real_type():
            const = self.manager.Real(constant_add)
        else:
            assert ttype.is_int_type()
            const = self.manager.Int(constant_add)

        if len(to_sum) == 0 and len(to_sub) == 0:
            return const
        if not const.is_zero():
            to_sum.append(const)

        assert to_sum or to_sub

        res = self.manager.Plus(to_sum) if to_sum else None

        if to_sub:
            sub = self.manager.Plus(to_sub)
            if res:
                res = self.manager.Minus(res, sub)
            else:
                if ttype.is_int_type():
                    m_1 = self.manager.Int(-1)
                else:
                    assert ttype.is_real_type()
                    m_1 = self.manager.Real(-1)
                res = self.manager.Times(m_1, sub)
        return res

    def walk_times(self, formula, args, **kwargs):
        new_args = []
        constant_mul = 1
        stack = list(args)
        ttype = self.env.stc.get_type(args[0])
        is_algebraic = False
        while len(stack) > 0:
            x = stack.pop()
            if x.is_constant():
                if x.is_algebraic_constant():
                    is_algebraic = True
                if x.is_zero():
                    constant_mul = 0
                    break
                else:
                    constant_mul *= x.constant_value()
            elif x.is_times():
                stack += x.args()
            else:
                new_args.append(x)

        const = None
        if is_algebraic:
            from pysmt.constants import Numeral
            const = self.manager._Algebraic(Numeral(constant_mul))
        elif ttype.is_real_type():
            const = self.manager.Real(constant_mul)
        else:
            assert ttype.is_int_type()
            const = self.manager.Int(constant_mul)

        if const.is_zero():
            return const
        else:
            if len(new_args) == 0:
                return const
            elif not const.is_one():
                new_args.append(const)

        new_args = sorted(new_args, key=FNode.node_id)
        return self.manager.Times(new_args)

    def walk_pow(self, formula, args, **kwargs):
        if args[0].is_real_constant():
            l = args[0].constant_value()
            r = args[1].constant_value()
            return self.manager.Real(l**r)

        if args[0].is_int_constant():
            l = args[0].constant_value()
            r = args[1].constant_value()
            return self.manager.Int(l**r)

        if args[0].is_algebraic_constant():
            from pysmt.constants import Numeral
            l = args[0].constant_value()
            r = args[1].constant_value()
            return self.manager._Algebraic(Numeral(l**r))

        return self.manager.Pow(args[0], args[1])

    def walk_minus(self, formula, args, **kwargs):
        assert len(args) == 2

        sl = args[0]
        sr = args[1]

        if sl.is_real_constant() and sr.is_real_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Real(l - r)

        if sl.is_int_constant() and sr.is_int_constant():
            l = sl.constant_value()
            r = sr.constant_value()
            return self.manager.Int(l - r)

        if sl.is_algebraic_constant() and \
           sr.is_algebraic_constant():
            from pysmt.constants import Numeral
            l = sl.constant_value()
            r = sr.constant_value()
            return self.mananger._Algebraic(Numeral(l - r))

        if sr.is_constant() and sr.is_zero():
            return sl

        if sl == sr:
            if self.env.stc.walk(sl) == types.REAL:
                return self.manager.Real(0)
            else:
                return self.manager.Int(0)

        return self.manager.Minus(sl, sr)

    def walk_function(self, formula, args, **kwargs):
        return self.manager.Function(formula.function_name(), args)

    def walk_toreal(self, formula, args, **kwargs):
        assert len(args) == 1
        if args[0].is_constant():
            assert args[0].is_int_constant()
            return self.manager.Real(args[0].constant_value())
        return self.manager.ToReal(args[0])

    def walk_bv_and(self, formula, args, **kwargs):
        simplified = None
        width = formula.bv_width()

        if args[0].is_bv_constant():
            lhs = args[0].bv_unsigned_value()

            if lhs == 0:
                # 0 & x -> 0
                simplified = self.manager.BVZero(width)
            elif lhs == 2**width - 1:
                # 0xf...f & x -> x
                simplified = args[1]
            elif args[1].is_bv_constant():
                rhs = args[1].bv_unsigned_value()
                res = lhs & rhs
                simplified = self.manager.BV(res, width=width)
        elif args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()

            if rhs == 0:
                # x & 0 -> 0
                simplified = self.manager.BVZero(width)
            elif rhs == 2**width - 1:
                # x & 0xf...f -> x
                simplified = args[0]

        if simplified is not None:
            return simplified

        return self.manager.BVAnd(args[0], args[1])

    def walk_bv_not(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            res = ~args[0].constant_value() & (2**formula.bv_width() - 1)
            return self.manager.BV(res, width=formula.bv_width())
        return self.manager.BVNot(args[0])

    def walk_bv_neg(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            res = 2**formula.bv_width() - args[0].constant_value()
            res = res % 2**formula.bv_width()
            return self.manager.BV(res, width=formula.bv_width())
        return self.manager.BVNeg(args[0])

    def walk_bv_or(self, formula, args, **kwargs):
        simplified = None

        if args[0].is_bv_constant():
            lhs = args[0].bv_unsigned_value()

            if lhs == 0:
                # 0 | x -> x
                simplified = args[1]
            else:
                width = formula.bv_width()
                mask = 2**width - 1

                if lhs == mask:
                    # 0xf...f | x -> 0xf...f
                    simplified = self.manager.BV(mask, width=width)
                elif args[1].is_constant():
                    res = lhs | args[1].bv_unsigned_value()
                    simplified = self.manager.BV(res, width=width)
        elif args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()

            if rhs == 0:
                # x | 0 -> x
                simplified = args[0]
            else:
                width = formula.bv_width()
                mask = 2**width - 1
                if rhs == mask:
                    # x | 0xf...f -> 0xf...f
                    simplified = self.manager.BV(mask, width=width)

        if simplified is not None:
            return simplified

        return self.manager.BVOr(args[0], args[1])

    def walk_bv_xor(self, formula, args, **kwargs):
        if args[0].is_bv_constant() and args[1].is_bv_constant():
            res = args[0].constant_value() ^ args[1].constant_value()
            return self.manager.BV(res, width=formula.bv_width())
        return self.manager.BVXor(args[0], args[1])

    def walk_bv_add(self, formula, args, **kwargs):
        simplified = None

        if args[0].is_bv_constant():
            lhs = args[0].bv_unsigned_value()
            if lhs == 0:
                # 0 + args[1] -> args[1]
                simplified = args[1]
            elif args[1].is_bv_constant():
                width = formula.bv_width()
                res = lhs + args[1].bv_unsigned_value()
                res = res % 2**width
                simplified = self.manager.BV(res, width=width)
        elif args[1].is_bv_constant() and args[1].bv_unsigned_value() == 0:
            # args[0] + 0 -> args[0]
            simplified = args[0]

        if simplified is not None:
            return simplified

        return self.manager.BVAdd(args[0], args[1])

    def walk_bv_mul(self, formula, args, **kwargs):
        simplified = None

        if args[0].is_bv_constant():
            lhs = args[0].bv_unsigned_value()
            if lhs == 0:
                # 0 * args[1] -> 0
                simplified = self.manager.BVZero(formula.bv_width())
            elif lhs == 1:
                # 1 * args[1] -> args[1]
                simplified = args[1]
            elif args[1].is_bv_constant():
                width = formula.bv_width()
                res = lhs * args[1].bv_unsigned_value()
                res = res % 2**width
                simplified = self.manager.BV(res, width=width)
        elif args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()
            if rhs == 0:
                # args[0] * 0 -> 0
                simplified = self.manager.BVZero(formula.bv_width())
            elif rhs == 1:
                # args[0] * 1 -> args[0]
                simplified = args[0]

        if simplified is not None:
            return simplified

        return self.manager.BVMul(args[0], args[1])

    def walk_bv_udiv(self, formula, args, **kwargs):
        simplified = None

        if args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()
            if rhs == 0:
                # args[0] / 0 -> all 1's
                width = formula.bv_width()
                simplified = self.manager.BV(2**width - 1, width=width)
            elif rhs == 1:
                # args[0] / 1 -> args[0]
                simplified = args[0]
            elif args[0].is_bv_constant():
                width = formula.bv_width()
                res = args[0].bv_unsigned_value() // rhs
                res = res % 2**width
                simplified = self.manager.BV(res, width=width)

        if simplified is not None:
            return simplified

        return self.manager.BVUDiv(args[0], args[1])

    def walk_bv_urem(self, formula, args, **kwargs):
        simplified = None

        if args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()
            if rhs == 0:
                # args[0] % 0 -> args[0]
                simplified = args[0]
            elif rhs == 1:
                # args[0] % 1 -> 0
                simplified = self.manager.BVZero(formula.bv_width())
            elif args[0].is_bv_constant():
                res = args[0].bv_unsigned_value() % rhs
                simplified = self.manager.BV(res, formula.bv_width())
        elif args[0].is_bv_constant() and args[0].bv_unsigned_value() == 0:
            # 0 % args[1] -> 0
            simplified = self.manager.BVZero(formula.bv_width())

        if simplified is not None:
            return simplified

        return self.manager.BVURem(args[0], args[1])

    def walk_bv_ult(self, formula, args, **kwargs):
        simplified = None

        if args[0] == args[1]:
            # x u< x -> FALSE
            simplified = self.manager.FALSE()
        elif args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()
            if rhs == 0:
                # args[0] u< 0 -> FALSE
                simplified = self.manager.FALSE()
            elif args[0].is_bv_constant():
                res = args[0].bv_unsigned_value() < rhs
                simplified = self.manager.Bool(res)

        if simplified is not None:
            return simplified

        return self.manager.BVULT(args[0], args[1])

    def walk_bv_ule(self, formula, args, **kwargs):
        simplified = None

        if args[0] == args[1]:
            # x u<= x -> TRUE
            simplified = self.manager.TRUE()
        elif args[0].is_bv_constant():
            lhs = args[0].bv_unsigned_value()
            if lhs == 0:
                # 0 u<= args[1] -> TRUE
                simplified = self.manager.TRUE()
            elif args[1].is_bv_constant():
                res = lhs <= args[1].bv_unsigned_value()
                simplified = self.manager.Bool(res)

        if simplified is not None:
            return simplified

        return self.manager.BVULE(args[0], args[1])

    def walk_bv_extract(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            bitstr = args[0].bv_bin_str(reverse=True)
            start = formula.bv_extract_start()
            end = formula.bv_extract_end()
            res = bitstr[start:end+1][::-1]
            w = (end+1)-start
            return self.manager.BV("#b%s" % res, width=w)
        return self.manager.BVExtract(args[0],
                                      start=formula.bv_extract_start(),
                                      end=formula.bv_extract_end())

    def walk_bv_ror(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            bitstr = args[0].bv_bin_str(reverse=True)
            # Takes first k elements and move to end
            slice1 = bitstr[0:formula.bv_rotation_step()]
            slice2 = bitstr[formula.bv_rotation_step():]
            res = (slice2 + slice1)[::-1]
            return self.manager.BV(res)
        return self.manager.BVRor(args[0], formula.bv_rotation_step())

    def walk_bv_rol(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            bitstr = args[0].bv_bin_str(reverse=True)
            # Takes last k elements and move to beginning
            slice1 = bitstr[0:-formula.bv_rotation_step()]
            slice2 = bitstr[-formula.bv_rotation_step():]
            res = (slice2 + slice1)[::-1]
            return self.manager.BV(res)
        return self.manager.BVRol(args[0], formula.bv_rotation_step())

    def walk_bv_sext(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            bitstr = args[0].bv_bin_str()
            filler = bitstr[0]
            res = filler*formula.bv_extend_step() + bitstr
            return self.manager.BV(res, width=formula.bv_width())
        return self.manager.BVSExt(args[0], formula.bv_extend_step())

    def walk_bv_zext(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            bitstr = args[0].bv_bin_str()
            filler = "0"
            res = filler*formula.bv_extend_step() + bitstr
            return self.manager.BV(res, width=formula.bv_width())
        return self.manager.BVZExt(args[0], formula.bv_extend_step())

    def walk_bv_concat(self, formula, args, **kwargs):
        if args[0].is_bv_constant() and args[1].is_bv_constant():
            w0 = args[0].bv_width()
            w1 = args[1].bv_width()
            res = (2**w1) * args[0].bv_unsigned_value() + \
                  args[1].bv_unsigned_value()
            return self.manager.BV(res, w1 + w0)
        return self.manager.BVConcat(args[0], args[1])

    def walk_bv_lshl(self, formula, args, **kwargs):
        simplified = None

        if args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()

            if rhs == 0:
                # x << 0 -> x
                simplified = args[0]
            else:
                width = args[0].bv_width()

                if rhs >= width:
                    # x_n << m -> 0, where m >= n
                    simplified = self.manager.BVZero(width)
                elif args[0].is_bv_constant():
                    res = args[0].bv_unsigned_value() << rhs
                    simplified = self.manager.BV(res % (2 ** width), width)
        elif args[0].is_bv_constant() and args[0].bv_unsigned_value() == 0:
            # 0 << x -> 0
            simplified = args[0]

        if simplified is not None:
            return simplified

        return self.manager.BVLShl(args[0], args[1])

    def walk_bv_lshr(self, formula, args, **kwargs):
        simplified = None

        if args[1].is_bv_constant():
            rhs = args[1].bv_unsigned_value()

            if rhs == 0:
                # x >> 0 -> x
                simplified = args[0]
            else:
                width = args[0].bv_width()

                if rhs >= width:
                    # x_n >> m -> 0, where m >= n
                    simplified = self.manager.BVZero(width)
                elif args[0].is_bv_constant():
                    res = args[0].bv_unsigned_value() >> rhs
                    simplified = self.manager.BV(res % (2 ** width), width)
        elif args[0].is_bv_constant() and args[0].bv_unsigned_value() == 0:
            # 0 >> x -> 0
            simplified = args[0]

        if simplified is not None:
            return simplified

        return self.manager.BVLShr(args[0], args[1])

    def walk_bv_sub(self, formula, args, **kwargs):
        simplified = None

        if args[0] == args[1]:
            # x - x -> 0
            simplified = self.manager.BVZero(formula.bv_width())
        if args[1].is_bv_constant():
            rhs = args[1].constant_value()

            if rhs == 0:
                # x - 0 -> x
                simplified = args[0]
            elif args[0].is_bv_constant():
                res = args[0].constant_value() - rhs
                res = res % 2**formula.bv_width()
                simplified = self.manager.BV(res, width=formula.bv_width())

        if simplified is not None:
            return simplified

        return self.manager.BVSub(args[0], args[1])

    def walk_bv_slt(self, formula, args, **kwargs):
        if args[0].is_bv_constant() and args[1].is_bv_constant():
            res = args[0].bv_signed_value() < args[1].bv_signed_value()
            return self.manager.Bool(res)
        if args[0] == args[1]:
            # x < x -> False
            return self.manager.Bool(False)
        return self.manager.BVSLT(args[0], args[1])

    def walk_bv_sle(self, formula, args, **kwargs):
        if args[0].is_bv_constant() and args[1].is_bv_constant():
            res = args[0].bv_signed_value() <= args[1].bv_signed_value()
            return self.manager.Bool(res)
        if args[0] == args[1]:
            # x <= x -> True
            return self.manager.Bool(True)
        return self.manager.BVSLE(args[0], args[1])

    def walk_bv_comp(self, formula, args, **kwargs):
        sl, sr = args

        if sl == sr:
            return self.manager.BV(1, 1)
        elif sl.is_bv_constant() and sr.is_bv_constant():
            return self.manager.BV(0, 1)
        else:
            return self.manager.BVComp(sl, sr)

    def walk_bv_sdiv(self, formula, args, **kwargs):
        l,r = args
        if l.is_bv_constant() and r.is_bv_constant():
            l_sign = l.bv_signed_value() < 0
            r_sign = r.bv_signed_value() < 0
            if (not l_sign) and (not r_sign):
                return self.walk_bv_udiv(self.manager.BVUDiv(l,r), args, **kwargs)
            elif l_sign and (not r_sign):
                nl = self.walk_bv_neg(self.manager.BVNeg(l), [l], **kwargs)
                div = self.walk_bv_udiv(self.manager.BVUDiv(nl, r), [nl, r],
                                        **kwargs)
                return self.walk_bv_neg(self.manager.BVNeg(div), [div], **kwargs)
            elif (not l_sign) and r_sign:
                nr = self.walk_bv_neg(self.manager.BVNeg(r), [r], **kwargs)
                div = self.walk_bv_udiv(self.manager.BVUDiv(l, nr), [l, nr],
                                        **kwargs)
                return self.walk_bv_neg(self.manager.BVNeg(div), [div], **kwargs)
            else:
                nl = self.walk_bv_neg(self.manager.BVNeg(l), [l], **kwargs)
                nr = self.walk_bv_neg(self.manager.BVNeg(r), [r], **kwargs)
                return self.walk_bv_udiv(self.manager.BVUDiv(nl, nr), [nl, nr],
                                         **kwargs)
        return self.manager.BVSDiv(l, r)

    def walk_bv_srem(self, formula, args, **kwargs):
        if args[0].is_bv_constant() and args[1].is_bv_constant():
            l = args[0]
            if args[0].bv_signed_value() < 0:
                l = self.walk_bv_neg(self.manager.BVNeg(args[0]), [args[0]],
                                     **kwargs)

            r = args[1]
            if args[1].bv_signed_value() < 0:
                r = self.walk_bv_neg(self.manager.BVNeg(args[1]), [args[1]],
                                     **kwargs)


            res = self.walk_bv_urem(self.manager.BVURem(l, r), [l, r],
                                    **kwargs)

            if args[0].bv_signed_value() < 0:
                res = self.walk_bv_neg(self.manager.BVNeg(res), [res],
                                       **kwargs)
            return res
        return self.manager.BVSRem(args[0], args[1])

    def walk_bv_ashr(self, formula, args, **kwargs):
        l,r = args
        if l.is_bv_constant() and r.is_bv_constant():
            sign = l.bv_signed_value() < 0
            ret = self.walk_bv_lshr(self.manager.BVLShr(l, r), [l, r], **kwargs)
            width = formula.bv_width()
            if sign:
                n = ret.bv_unsigned_value()
                padlen = width
                if width > r.bv_unsigned_value():
                    padlen = r.bv_unsigned_value()

                for i in range(width-padlen, width):
                    n = set_bit(n, i, True)
                ret = self.manager.BV(n, width)
            return ret
        return self.manager.BVAShr(l, r)

    def walk_str_length(self, formula, args, **kwargs):
        s = args[0]
        if s.is_string_constant():
            return self.manager.Int(len(s.constant_value()))
        return self.manager.StrLength(s)

    def walk_str_concat(self, formula, args, **kwargs):
        if all(arg.is_string_constant() for arg in args):
            ret = "".join(x.constant_value() for x in args)
            return self.manager.String(ret)
        return self.manager.StrConcat(args)

    def walk_str_charat(self, formula,  args, **kwargs):
        s, i = args
        if s.is_string_constant() and i.is_int_constant():
            res = s.constant_value()[i.constant_value():i.constant_value() + 1]
            return self.manager.String(res)
        return self.manager.StrCharat(s, i)

    def walk_str_contains(self, formula, args, **kwargs):
        s,t = args
        if s.is_string_constant() and t.is_string_constant():
            return self.manager.Bool(t.constant_value() in s.constant_value())
        return self.manager.StrContains(s, t)

    def walk_str_indexof(self, formula, args, **kwargs):
        s, t, i = args
        if s.is_string_constant() and t.is_string_constant() and i.is_int_constant():
            idx = s.constant_value().find(t.constant_value(), i.constant_value())
            # idx = -1, if t is not found
            return self.manager.Int(idx)
        return self.manager.StrIndexOf(s, t, i)

    def walk_str_replace(self, formula, args, **kwargs):
        s, t1, t2 = args
        if s.is_string_constant() and t1.is_string_constant() and t2.is_string_constant():
            t1_str = t1.constant_value()
            t2_str = t2.constant_value()
            s_str = s.constant_value()
            return self.manager.String(s_str.replace(t1_str, t2_str, 1))
        return self.manager.StrReplace(s, t1, t2)

    def walk_str_substr(self, formula, args, **kwargs):
        s, i, j = args
        if s.is_string_constant() and i.is_int_constant() and j.is_int_constant():
            start_ = i.constant_value()
            end_ = i.constant_value() + j.constant_value()
            res = s.constant_value()[start_:end_]
            return self.manager.String(res)
        return self.manager.StrSubstr(s, i, j)

    def walk_str_prefixof(self, formula, args, **kwargs):
        s, t = args
        if s.is_string_constant() and t.is_string_constant():
            return self.manager.Bool(t.constant_value().startswith(s.constant_value()))
        return self.manager.StrPrefixof(s, t)

    def walk_str_suffixof(self, formula, args, **kwargs):
        s, t = args
        if s.is_string_constant() and t.is_string_constant():
            return self.manager.Bool(t.constant_value().endswith(s.constant_value()))
        return self.manager.StrSuffixof(s, t)

    def walk_str_to_int(self, formula, args, **kwargs):
        s = args[0]
        if s.is_string_constant():
            try:
                return self.manager.Int(int(s.constant_value()))
            except ValueError:
                return self.manager.Int(-1)
        return self.manager.StrToInt(s)

    def walk_int_to_str(self, formula, args, **kwargs):
        i = args[0]
        if i.is_int_constant():
            if i.constant_value() < 0:
                return self.manager.String("")
            return self.manager.String(str(i.constant_value()))
        return self.manager.IntToStr(i)

    def walk_bv_tonatural(self, formula, args, **kwargs):
        if args[0].is_bv_constant():
            return self.manager.Int(args[0].constant_value())
        return self.manager.BVToNatural(args[0])

    def walk_array_select(self, formula, args, **kwargs):
        a, i = args
        if a.is_array_value() and i.is_constant():
            return a.array_value_get(i)
        return self.manager.Select(args[0], args[1])

    def walk_array_store(self, formula, args, **kwargs):
        a, i, v = args
        if a.is_array_value() and i.is_constant():
            assign = a.array_value_assigned_values_map()
            assign[i] = v # Add / Overwrite assignment at index i
            return self.manager.Array(a.array_value_index_type(),
                                      a.array_value_default(),
                                      assign)
        return self.manager.Store(a, i, v)

    def walk_array_value(self, formula, args, **kwargs):
        assign = dict(zip(args[1::2], args[2::2]))
        return self.manager.Array(formula.array_value_index_type(),
                                  args[0],
                                  assign)

    def walk_div(self, formula, args, **kwargs):
        sl = args[0]
        sr = args[1]

        if sl.is_constant() and sr.is_constant() and not sr.is_zero():
            l = sl.constant_value()
            r = sr.constant_value()
            if sl.is_real_constant():
                return self.manager.Real(l / r)
            else:
                assert sl.is_int_constant()
                # smtlib2 semantics of integer division:
                # r > 0 : l / r == floor(float(l) / r)
                # r < 0 : l / r == ceil(float(l) / r)
                if r > 0:
                    return self.manager.Int(math.floor(float(l) / r))
                if r < 0:
                    return self.manager.Int(math.ceil(float(l) / r))

        if sl.is_constant():
            if sl.is_zero():
                return sl

        if sr.is_constant():
            if sr.is_one():
                return sl

        return self.manager.Div(sl, sr)

    @handles(op.SYMBOL)
    @handles(op.REAL_CONSTANT, op.INT_CONSTANT, op.BOOL_CONSTANT)
    @handles(op.BV_CONSTANT, op.STR_CONSTANT, op.ALGEBRAIC_CONSTANT)
    def walk_identity(self, formula, args, **kwargs):
        return formula

# EOC Simplifier


class BddSimplifier(Simplifier):
    """A simplifier relying on BDDs.

    The formula is translated into a BDD and then translated back into
    a pySMT formula. This is a much more expensive simplification
    process, and might not work with formulas with thousands of
    boolean variables.

    The option ``static_ordering`` can be used to provide a variable
    ordering for the underlying bdd.

    The option ``bool_abstraction`` controls how to behave if the
    input formula contains Theory terms (i.e., is not purely boolean).
    If this option is False (default) an exception will be thrown when
    a Theory atom is found. If it is set to True, the Theory part is
    abstracted, and the simplification is performed only on the
    boolean structure of the formula.
    """

    def __init__(self, env=None, static_ordering=None, bool_abstraction=False):
        Simplifier.__init__(self, env=env)
        self._validation_sname = None

        Solver = self.env.factory.Solver
        if static_ordering is not None:
            solver_options={'static_ordering': static_ordering}
        else:
            solver_options={'dynamic_reordering': True}
        self.s = Solver(name="bdd", solver_options=solver_options)
        self.convert = self.s.converter.convert
        self.back = self.s.converter.back
        # Set methods for boolean_abstraction
        self.bool_abstraction = bool_abstraction
        self.ba_map = {}
        self.get_type = self.env.stc.get_type
        self.FreshSymbol = self.env.formula_manager.FreshSymbol

    @property
    def validate_simplifications(self):
        return self._validate_simplifications

    @validate_simplifications.setter
    def validate_simplifications(self, value):
        possible_solvers = [sname for sname in self.env.factory.all_solvers()\
                            if sname!="bdd"]
        if len(possible_solvers) == 0:
            raise PysmtValueError("To validate at least another solver must "
                                  "be available!")
        self._validation_sname = possible_solvers[0]
        self._validate_simplifications = value

    def simplify(self, formula):
        from pysmt.oracles import get_logic
        from pysmt.logics import BOOL, QF_BOOL
        if self.bool_abstraction:
            logic = get_logic(formula)
            if logic > QF_BOOL and logic != BOOL:
                res = self.abstract_and_simplify(formula)
            else:
                res = self.back(self.convert(formula))
        else:
            res = self.back(self.convert(formula))
        self._validate(formula, res)
        return res

    def _validate(self, old, new):
        if self.validate_simplifications:
            Iff = self.env.formula_manager.Iff
            is_valid = self.env.factory.is_valid
            sname = self._validation_sname
            assert is_valid(Iff(old, new), solver_name=sname ), \
              "Was: %s \n Obtained: %s\n" % (str(old), str(new))

    def abstract_and_simplify(self, formula):
        abs_formula = self.walk(formula)
        abs_res = self.back(self.convert(abs_formula))
        res = abs_res.substitute(self.ba_map)
        return res

    @handles(op.RELATIONS)
    def walk_simplify_and_abstract(self, formula, args, **kwargs):
        rewritten = Simplifier.super(self, formula, args, **kwargs)
        if rewritten.is_bool_constant():
            return rewritten
        new_var = self.FreshSymbol()
        self.ba_map[new_var] = rewritten
        return new_var

    def walk_function(self, formula, args, **kwargs):
        rewritten = Simplifier.walk_function(self, formula, args, **kwargs)
        if rewritten.function_name().symbol_type().return_type.is_bool_type():
            new_var = self.FreshSymbol()
            self.ba_map[new_var] = rewritten
            return new_var
        return rewritten


#EOC BddSimplifier



# This example requires Z3
#
# This example shows how to parallelize solving using pySMT, by using
# the multiprocessing library.
#
# All shortcuts (is_sat, is_unsat, get_model, etc.) can be easily
# used. We show two techniques: map and apply_async.
#
# Map applies a given function to a list of elements, and returns a
# list of results. We show an example using is_sat.
#
# Apply_async provides a way to perform operations in an asynchronous
# way. This allows us to start multiple solving processes in parallel,
# and collect the results whenever they become available.
#
# More information can be found in the Python Standard Library
# documentation for the multiprocessing module.
#
#
# NOTE: When running this example, a warning will appear saying:
#   "Contextualizing formula during is_sat"
#
# Each process will inherit a different formula manager, but the
# formulas that we are solving have been all created in the same
# formula manager. pySMT takes care of moving the formula across
# formula managers. This step is called contextualization, and occurs
# only when using multiprocessing.
# To disable the warning see the python module warnings.
#
from multiprocessing import Pool, TimeoutError
from time import sleep

from pysmt.test.examples import get_example_formulae
from pysmt.shortcuts import is_sat, is_valid, is_unsat
from pysmt.shortcuts import And

# Ignore this for now
def check_validity_and_test(args):
    """Checks expression and compare the outcome against a known value."""

    expr, expected = args # IMPORTANT: Unpack args !!!
    local_res = is_valid(expr)
    return local_res == expected


# Create the Pool with 4 workers.
pool = Pool(4)

# We use the examples formula from the test suite.
# This generator iterates over all the expressions
f_gen = (f.expr for f in get_example_formulae())

# Call the function is_sat on each expression
res = pool.map(is_sat, f_gen)

# The result is a list of True/False, in the same order as the input.
print(res)

sleep(1) # Have some time to look at the result

# Notice that all shortcuts (is_sat, qelim, etc) require only one
# mandatory argument. To pass multiple arguments, we need to pack them
# together.

# Lets create a list of (formula, result), where result is the
# expected result of the is_valid query.
f_gen = (f.expr for f in get_example_formulae())
res_gen = (f.is_valid for f in get_example_formulae())
args_gen = zip(f_gen, res_gen)

# We now define a function that check the formula against the expected
# value of is_valid: See check_validity_and_test(...) above.
# Due to the way multiprocessing works, we need to define the function
# _before_ we create the Pool.

# As before, we call the map on the pool
res = pool.map(check_validity_and_test, args_gen)

# We can also apply a map-reduce strategy, by making sure that all
# results are as expected.
if all(res):
    print("Everything is ok!")
else:
    print("Ooops, something is wrong!")
    print(res)

# A different option is to run solvers in an asynchronous way.  This
# does not provide us with any guarantee on the execution order, but
# it is particular convenient, if we want to perform solving together
# with another long operation (e.g., I/O, network etc.)  or if we want
# to run multiple solvers.

# Create a formula
big_f = And(f.expr for f in get_example_formulae() \
            if not f.logic.theory.bit_vectors and \
               not f.logic.theory.arrays and \
               not f.logic.theory.strings and \
                   f.logic.theory.linear)

# Create keyword arguments for the function call.
# This is the simplest way to pass multiple arguments to apply_async.
kwargs = {"formula": big_f, "solver_name": "z3"}
future_res_sat = pool.apply_async(is_sat, kwds=kwargs)
future_res_unsat = pool.apply_async(is_unsat, kwds=kwargs)

# In the background, the solving is taking place... We can do other
# stuff in the meanwhile.
print("This is non-blocking...")

# Get the result with a deadline.
# See multiprocessing.pool.AsyncResult for more options
sat_res = future_res_sat.get(10)  # Get result after 10 seconds or kill
try:
    unsat_res = future_res_unsat.get(0) # No wait
except TimeoutError:
    print("UNSAT result was not ready!")
    unsat_res = None
print(sat_res, unsat_res)



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import collections
from io import StringIO

from pysmt.test import TestCase, main
from pysmt.smtlib.parser import SmtLibParser
from pysmt.exceptions import UnknownSmtLibCommandError, PysmtValueError

TS = collections.namedtuple('TS', ['init', 'trans'])
TSFormula = collections.namedtuple('TSFormula', ['formula', "is_init"])

class TSSmtLibParser(SmtLibParser):
    def __init__(self, env=None, interactive=False):
        SmtLibParser.__init__(self, env, interactive)

        # Add new commands
        self.commands["init"] = self._cmd_init
        self.commands["trans"] = self._cmd_trans

        # Remove unused commands
        del self.commands["check-sat"]
        del self.commands["get-value"]
        # ...

        # Add 'next' function
        self.interpreted["next"] = self._operator_adapter(self._next_var)

    def _cmd_init(self, current, tokens):
        expr = self.get_expression(tokens)
        self.consume_closing(tokens, current)
        return TSFormula(expr, True)

    def _cmd_trans(self, current, tokens):
        expr = self.get_expression(tokens)
        self.consume_closing(tokens, current)
        return TSFormula(expr, False)

    def _next_var(self, symbol):
        if symbol.is_symbol():
            name = symbol.symbol_name()
            ty = symbol.symbol_type()
            return self.env.formula_manager.Symbol("next_" + name, ty)
        else:
            raise PysmtValueError("'next' operator can be applied only to symbols")

    def get_ts(self, script):
        init = self.env.formula_manager.TRUE()
        trans = self.env.formula_manager.TRUE()

        for cmd in self.get_command_generator(script):
            # Simply skip declarations and other commands...
            if type(cmd) == TSFormula:
                if cmd.is_init:
                    init = cmd.formula
                else:
                    trans = cmd.formula

        return TS(init, trans)

    def get_ts_fname(self, ts_fname):
        with open(ts_fname, 'r') as script:
            return self.get_ts(script)


class TestParserExtensibility(TestCase):

    def setUp(self):
        self.ts_parser = TSSmtLibParser()
        self.smt_parser = SmtLibParser()

    def test_wrong(self):
        txt = """
        (declare-fun A () Bool)
        (declare-fun B () Bool)
        (assert (and A B))
        (check-sat)
        (exit)
        """
        with self.assertRaises(UnknownSmtLibCommandError):
            self.ts_parser.get_ts(StringIO(txt))
        script = self.smt_parser.get_script(StringIO(txt))
        self.assertIsNotNone(script)

    def test_basic(self):
        txt = """
        (declare-fun A () Bool)
        (declare-fun B () Bool)
        (init (and A B))
        (trans (=> A (next A)))
        (exit)
        """
        with self.assertRaises(UnknownSmtLibCommandError):
            self.smt_parser.get_script(StringIO(txt))
        ts = self.ts_parser.get_ts(StringIO(txt))
        self.assertIsNotNone(ts)


if __name__ == '__main__':
    main()



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
The Environment is a key structure in pySMT. It contains multiple
singleton objects that are used throughout the system, such as the
FormulaManager, Simplifier, HRSerializer, SimpleTypeChecker.
"""

import pysmt.simplifier
import pysmt.printers
import pysmt.substituter
import pysmt.type_checker
import pysmt.oracles
import pysmt.formula
import pysmt.factory
import pysmt.decorators
import pysmt.typing


class Environment(object):
    """The Environment provides global singleton instances of various objects.

    FormulaManager and the TypeChecker are among the most commonly used ones.

    Subclasses of Environment should take care of adjusting the list
    of classes for the different services, by changing the class
    attributes.
    """
    TypeCheckerClass = pysmt.type_checker.SimpleTypeChecker
    FormulaManagerClass = pysmt.formula.FormulaManager
    TypeManagerClass = pysmt.typing.TypeManager
    SimplifierClass = pysmt.simplifier.Simplifier
    #SubstituterClass = pysmt.substituter.MSSubstituter
    SubstituterClass = pysmt.substituter.MGSubstituter
    HRSerializerClass = pysmt.printers.HRSerializer
    QuantifierOracleClass = pysmt.oracles.QuantifierOracle
    TheoryOracleClass = pysmt.oracles.TheoryOracle
    FreeVarsOracleClass= pysmt.oracles.FreeVarsOracle
    SizeOracleClass = pysmt.oracles.SizeOracle
    AtomsOracleClass = pysmt.oracles.AtomsOracle
    TypesOracleClass = pysmt.oracles.TypesOracle

    def __init__(self):
        self._stc = self.TypeCheckerClass(self)
        self._formula_manager = self.FormulaManagerClass(self)
        # NOTE: Both Simplifier and Substituter keep an internal copy
        # of the Formula Manager and need to be initialized afterwards
        self._simplifier = self.SimplifierClass(self)
        self._substituter = self.SubstituterClass(self)
        self._serializer = self.HRSerializerClass(self)
        self._qfo = self.QuantifierOracleClass(self)
        self._theoryo = self.TheoryOracleClass(self)
        self._fvo = self.FreeVarsOracleClass(self)
        self._sizeo = self.SizeOracleClass(self)
        self._ao = self.AtomsOracleClass(self)
        self._typeso = self.TypesOracleClass(self)
        self._type_manager = self.TypeManagerClass(self)

        self._factory = None
        # Configurations
        self.enable_infix_notation = False
        self.enable_div_by_0 = True

        # This option allows the construction of a symbol with empty
        # name (i.e. `Symbol("", INT)`). This feature is allowed by
        # SmtLib, but not all solvers support this and the machinery
        # needed to rename this one symbol is simply not worth the
        # effort given that such a name is simply bad
        # practice. However to validate pathological models we can
        # allow this on the pysmt side to use walkers
        # (e.g. substitution and simplification).
        self.allow_empty_var_names = False

        # Dynamic Walker Configuration Map
        # See: add_dynamic_walker_function
        self.dwf = {}

    @property
    def formula_manager(self):
        return self._formula_manager

    @property
    def type_manager(self):
        return self._type_manager

    @property
    def simplifier(self):
        return self._simplifier

    @property
    def serializer(self):
        return self._serializer

    @property
    def substituter(self):
        return self._substituter

    @property
    def stc(self):
        """ Get the Simple Type Checker """
        return self._stc

    @property
    def qfo(self):
        """ Get the Quantifier Oracle """
        return self._qfo

    @property
    def ao(self):
        """ Get the Atoms Oracle """
        return self._ao

    @property
    def theoryo(self):
        """ Get the Theory Oracle """
        return self._theoryo

    @property
    def typeso(self):
        """ Get the Types Oracle """
        return self._typeso

    @property
    def fvo(self):
        """ Get the FreeVars Oracle """
        return self._fvo

    @property
    def sizeo(self):
        """ Get the Size Oracle """
        return self._sizeo

    def add_dynamic_walker_function(self, nodetype, walker, function):
        """Dynamically bind the given function to the walker for the nodetype.

        This function enables the extension of walkers for new
        nodetypes. When introducing a new nodetype, we link a new
        function to a given walker, so that the walker will be able to
        handle the new nodetype.

        See :py:meth:`pysmt.walkers.generic.Walker.walk_error` for
        more information.
        """
        # self.dwf is a map of maps: {nodetype, {walker: function}}
        if nodetype not in self.dwf:
            self.dwf[nodetype] = {}

        assert walker not in self.dwf[nodetype], "Redefinition"
        self.dwf[nodetype][walker] = function

    @property
    def factory(self):
        if self._factory is None:
            self._factory = pysmt.factory.Factory(self)
        return self._factory

    def __enter__(self):
        """Entering a Context """
        push_env(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Remove environment from global stack."""
        pop_env()

# EOC Environment

#### GLOBAL ENVIRONMENTS STACKS ####
ENVIRONMENTS_STACK = []

def get_env():
    """Returns the Environment at the head of the stack."""
    return ENVIRONMENTS_STACK[-1]

def push_env(env=None):
    """Push a env in the stack. If env is None, a new Environment is created."""
    if env is None:
        env = Environment()
    ENVIRONMENTS_STACK.append(env)

def pop_env():
    """Pop an env from the stack."""
    return ENVIRONMENTS_STACK.pop()

def reset_env():
    """Destroys and recreate the head environment."""
    pop_env()
    push_env()
    return get_env()

# Create the default environment
push_env()



# This is a different take on the puzzle.py example
#
# This examples shows how to:
# 1. Enable and use infix notation
# 2. Use a solver context
#
from pysmt.shortcuts import Symbol, And, Plus, Int
from pysmt.shortcuts import Solver
from pysmt.typing import INT

# Infix-Notation is automatically enabled whenever you import pysmt.shortcuts.
#
# To enable it without using shortcuts, do:
#
#   from pysmt.environment import get_env
#   get_env().enable_infix_notation = True
#
# Similarly, you can disable infix_notation to prevent its accidental use.
#

hello = [Symbol(s, INT) for s in "hello"]
world = [Symbol(s, INT) for s in "world"]
letters = set(hello+world)

# Infix notation for Theory atoms does the overloading of python
# operator. For boolean connectors, we use e.g., x.And(y) This
# increases readability without running into problems of operator
# precedence.
#
# Note how you can mix prefix and infix notation.
domains = And([(Int(1) <= l).And(Int(10) >= l) for l in letters])

sum_hello = Plus(hello) # n-ary operators can take lists
sum_world = Plus(world) # as arguments
problem = (sum_hello.Equals(sum_world)).And(sum_hello.Equals(Int(25)))
formula = domains.And(problem)

print("Serialization of the formula:")
print(formula)

# A context (with-statment) lets python take care of creating and
# destroying the solver.
with Solver() as solver:
    solver.add_assertion(formula)
    if solver.solve():
        for l in letters:
            print("%s = %s" %(l, solver.get_value(l)))
    else:
        print("No solution found")
